import numpy as np
import pandas as pd
import scipy.stats
import umap
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE


def calculate_coefficient_variation(values: np.ndarray) -> np.ndarray:
    """
    Compute the coefficient of variation, the ratio of the biased standard
    deviation to the mean, in percentage. For more information
    visit https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.variation.html.

    :param numpy.ndarray values: numpy array of log2 transformed values
    :return: The calculated variation along rows.
    :rtype: numpy.ndarray

    Example::

        result = calculate_coefficient_variation()
    """
    cv = scipy.stats.variation(values.apply(lambda x: np.power(2, x)).values) * 100

    return cv


def get_coefficient_variation(data, drop_columns, group, columns=["name", "y"]):
    """
    Extracts the coefficients of variation in each group.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns
                 (with additional columns 'group', 'sample' and 'subject').
    :param list drop_columns: column labels to be dropped from the dataframe
    :param str group: column label containing group identifiers.
    :param list columns: names to use for the variable column(s), and for the value column(s)
    :return: Pandas dataframe with columns 'name' (protein identifier),
             'x' (coefficient of variation), 'y' (mean) and 'group'.

    Example::

        result = get_coefficient_variation(data, drop_columns=['sample', 'subject'], group='group')
    """
    df = data.copy()
    formated_df = df.drop(drop_columns, axis=1)
    cvs = formated_df.groupby(group).apply(func=calculate_coefficient_variation)
    cols = formated_df.set_index(group).columns.tolist()
    cvs_df = pd.DataFrame()
    for i in cvs.index:
        gcvs = cvs[i].tolist()
        ints = formated_df.set_index(group).mean().values.tolist()
        tdf = pd.DataFrame(data={"name": cols, "x": gcvs, "y": ints})
        tdf[group] = i

        if cvs_df.empty:
            cvs_df = tdf.copy()
        else:
            cvs_df = pd.concat([cvs_df, tdf])

    return cvs_df


def extract_number_missing(data, min_valid, drop_cols=["sample"], group="group"):
    """
    Counts how many valid values exist in each column and filters column labels with more
    valid values than the minimum threshold defined.

    :param data: pandas DataFrame with group as rows and protein identifier as column.
    :param str group: column label containing group identifiers.
                      If None, number of valid values is counted across all samples,
                      otherwise is counted per unique group identifier.
    :param int min_valid: minimum number of valid values to be filtered.
    :param list drop_columns: column labels to be dropped.
    :return: List of column labels above the threshold.

    Example::

        result = extract_number_missing(data, min_valid=3, drop_cols=['sample'], group='group')
    """
    if group is None:
        groups = data.loc[:, data.notnull().sum(axis=0) >= min_valid]
    else:
        groups = data.copy()
        if len(set(drop_cols).intersection(groups.columns.tolist())) == len(drop_cols):
            groups = groups.drop(drop_cols, axis=1)
        groups = groups.set_index(group).notnull().groupby(level=0).sum()
        groups = groups[groups >= min_valid]

    groups = groups.dropna(how="all", axis=1)
    return groups.columns.unique().tolist()


def extract_percentage_missing(
    data, missing_max, drop_cols=["sample"], group="group", how="all"
):
    """
    Extracts ratio of missing/valid values in each column and filters column labels with
    lower ratio than the minimum threshold defined.

    :param data: pandas dataframe with group as rows and protein identifier as column.
    :param str group: column label containing group identifiers.
                      If None, ratio is calculated across all samples,
                      otherwise is calculated per unique group identifier.
    :param float missing_max: maximum ratio of missing/valid values to be filtered.
    :param str how: define if labels with a higher percentage of missing values than the threshold
                    in any group ('any') or in all groups ('all') should be filtered
    :return: List of column labels below the threshold.

    Example::
        result = extract_percentage_missing(data, missing_max=0.3,
                                            drop_cols=['sample'], group='group')
    """
    if group is None:
        groups = data.loc[:, data.isnull().mean() <= missing_max].columns
    else:
        groups = data.copy()
        groups = groups.drop(drop_cols, axis=1)
        groups = groups.set_index(group)
        groups = groups.isnull().groupby(level=0).mean()
        groups = groups[groups <= missing_max]
        if how == "all":
            groups = groups.dropna(how="all", axis=1).columns.unique().tolist()
        elif how == "any":
            groups = groups.dropna(how="any", axis=1).columns.unique().tolist()
        else:
            if how in groups.index:
                groups = groups.loc[how, :].dropna().index.unique().tolist()

    return groups


def run_pca(
    data,
    drop_cols=["sample", "subject"],
    group="group",
    annotation_cols=["sample"],
    components=2,
    dropna=True,
):
    """
    Performs principal component analysis and returns the values of each component for each sample
    and each protein, and the loadings for each protein.

    For information visit
    https://scikit-learn.org/stable/modules/generated/sklearn.decomposition.PCA.html.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns
                 (with additional columns 'group', 'sample' and 'subject').
    :param list drop_cols: column labels to be dropped from the dataframe.
    :param str group: column label containing group identifiers.
    :param list annotation_cols: list of columns to be added in the scatter plot annotation
    :param int components: number of components to keep.
    :param bool dropna: if True removes all columns with any missing values.
    :return: tuple: 1) three pandas dataframes: components, loadings and variance; 2)
             xaxis and yaxis titles with components loadings for plotly.

    Example::

        result = run_pca(data, drop_cols=['sample', 'subject'], group='group',
                         components=2, dropna=True)
    """

    np.random.seed(112736)
    var_exp = []
    args = {}
    if data.empty:
        raise ValueError("Dataframe is empty.")

    df = data.copy()
    annotations = pd.DataFrame()
    if annotation_cols is not None:
        if len(list(set(annotation_cols).intersection(data.columns))) > 0:
            annotations = data.set_index(group)[annotation_cols]
    drop_cols_int = list(set(drop_cols).intersection(df.columns))
    if len(drop_cols_int) > 0:
        df = df.drop(drop_cols_int, axis=1)

    y = df[group].tolist()
    df = df.set_index(group)
    df = df.select_dtypes(["number"])
    if dropna:
        df = df.dropna(axis=1)
    X = df.values

    if X.size > 0 and X.shape[1] > components:
        pca = PCA(n_components=components)
        X = pca.fit_transform(X)
        var_exp = pca.explained_variance_ratio_
        loadings = pd.DataFrame(pca.components_.transpose())
        loadings.index = df.columns
        values = {
            index: np.sqrt(np.power(row, 2).sum()) for index, row in loadings.iterrows()
        }
        loadings["value"] = loadings.index.map(values.get)
        loadings = loadings.sort_values(by="value", ascending=False)
        args = {
            "x_title": "PC1" + " ({0:.2f})".format(var_exp[0]),
            "y_title": "PC2" + " ({0:.2f})".format(var_exp[1]),
            "group": "group",
        }
        if components == 2:
            resultDf = pd.DataFrame(X, index=y, columns=["x", "y"])
            resultDf = resultDf.assign(**annotations)
            resultDf = resultDf.reset_index()
            resultDf.columns = ["group", "x", "y"] + annotation_cols

            loadings.columns = ["x", "y", "value"]
        if components > 2:
            args.update({"z_title": "PC3" + " ({0:.2f})".format(var_exp[2])})
            resultDf = pd.DataFrame(X, index=y)
            resultDf = resultDf.assign(**annotations)
            resultDf = resultDf.reset_index()
            pca_cols = []
            loading_cols = []
            if components > 3:
                pca_cols = [str(i) for i in resultDf.columns[4:]]
                loading_cols = [str(i) for i in loadings.columns[3:]]

            resultDf.columns = ["group", "x", "y", "z"] + pca_cols
            loadings.columns = ["x", "y", "z"] + loading_cols

    return (resultDf, loadings, var_exp), args


def run_tsne(
    data,
    drop_cols=["sample", "subject"],
    group="group",
    annotation_cols=["sample"],
    components=2,
    perplexity=40,
    n_iter=1000,
    init="pca",
    dropna=True,
):
    """
    Performs t-distributed Stochastic Neighbor Embedding analysis.

    For more information visit
    https://scikit-learn.org/stable/modules/generated/sklearn.manifold.TSNE.html.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns
                 (with additional columns 'group', 'sample' and 'subject').
    :param list drop_cols: column labels to be dropped from the dataframe.
    :param str group: column label containing group identifiers.
    :param int components: dimension of the embedded space.
    :param list annotation_cols: list of columns to be added in the scatter plot annotation
    :param int perplexity: related to the number of nearest neighbors that is used
                           in other manifold learning algorithms.
                           Consider selecting a value between 5 and 50.
    :param int n_iter: maximum number of iterations for the optimization (at least 250).
    :param str init: initialization of embedding ('random', 'pca' or
                     numpy array of shape n_samples x n_components).
    :param bool dropna: if True removes all columns with any missing values.
    :return: Two dictionaries:
                1) pandas dataframe with embedding vectors,
                2) xaxis and yaxis titles for plotly.

    Example::

        result = run_tsne(data,
                          drop_cols=['sample', 'subject'],
                          group='group',
                          components=2,
                          perplexity=40,
                          n_iter=1000,
                          init='pca',
                          dropna=True
                        )
    """
    result = {}
    args = {}
    df = data.copy()
    if len(set(drop_cols).intersection(df.columns)) == len(drop_cols):
        df = df.drop(drop_cols, axis=1)
    df = df.set_index(group)
    if dropna:
        df = df.dropna(axis=1)
    df = df.select_dtypes(["number"])
    X = df.values
    y = df.index
    annotations = pd.DataFrame()
    if annotation_cols is not None:
        if len(list(set(annotation_cols).intersection(data.columns))) > 0:
            annotations = data[annotation_cols]
    if X.size > 0:
        tsne = TSNE(
            n_components=components,
            verbose=0,
            perplexity=perplexity,
            n_iter=n_iter,
            init=init,
        )
        X = tsne.fit_transform(X)
        args = {"x_title": "C1", "y_title": "C2"}
        if components == 2:
            resultDf = pd.DataFrame(X, index=y, columns=["x", "y"])
            resultDf = resultDf.reset_index()
            resultDf.columns = ["group", "x", "y"]
        if components > 2:
            args.update({"z_title": "C3"})
            resultDf = pd.DataFrame(X, index=y)
            resultDf = resultDf.reset_index()
            cols = []
            if len(components) > 4:
                cols = resultDf.columns[4:]
            resultDf.columns = ["group", "x", "y", "z"] + cols
        resultDf = resultDf.join(annotations)
        result["tsne"] = resultDf
    return result, args


def run_umap(
    data,
    drop_cols=["sample", "subject"],
    group="group",
    annotation_cols=["sample"],
    n_neighbors=10,
    min_dist=0.3,
    metric="cosine",
    dropna=True,
):
    """
    Performs Uniform Manifold Approximation and Projection.

    For more information vist https://umap-learn.readthedocs.io.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns
                 (with additional columns 'group', 'sample' and 'subject').
    :param list drop_cols: column labels to be dropped from the dataframe.
    :param str group: column label containing group identifiers.
    :param list annotation_cols: list of columns to be added in the scatter plot annotation
    :param int n_neighbors: number of neighboring points used
                            in local approximations of manifold structure.
    :param float min_dist: controls how tightly the embedding is allowed compress points together.
    :param str metric: metric used to measure distance in the input space.
    :param bool dropna: if True removes all columns with any missing values.
    :return: Two dictionaries:
                1) pandas dataframe with embedding of the training data in low-dimensional space,
                2) xaxis and yaxis titles for plotly.

    Example::

        result = run_umap(data,
                          drop_cols=['sample', 'subject'],
                          group='group',
                          n_neighbors=10,
                          min_dist=0.3,
                          metric='cosine',
                          dropna=True
                        )
    """
    np.random.seed(1145536)
    result = {}
    args = {}
    df = data.copy()
    if len(set(drop_cols).intersection(df.columns)) == len(drop_cols):
        df = df.drop(drop_cols, axis=1)
    df = df.set_index(group)
    if dropna:
        df = df.dropna(axis=1)
    df = df.select_dtypes(["number"])
    X = df.values
    y = df.index

    annotations = pd.DataFrame()
    if annotation_cols is not None:
        if len(list(set(annotation_cols).intersection(data.columns))) > 0:
            annotations = data[annotation_cols]

    if X.size:
        X = umap.UMAP(
            n_neighbors=n_neighbors, min_dist=min_dist, metric=metric
        ).fit_transform(X)
        args = {"x_title": "C1", "y_title": "C2"}
        resultDf = pd.DataFrame(X, index=y)
        resultDf = resultDf.reset_index()
        cols = []
        if len(resultDf.columns) > 3:
            cols = resultDf.columns[3:]
        resultDf.columns = ["group", "x", "y"] + cols
        resultDf = resultDf.join(annotations)
        result["umap"] = resultDf

    return result, args



import itertools

import numpy as np
import pandas as pd
import pingouin as pg
from scipy import stats
from scipy.special import betainc

import acore.utils as utils
from acore.multiple_testing import apply_pvalue_correction


def corr_lower_triangle(df: pd.DataFrame, **kwargs) -> pd.DataFrame:
    """Compute the correlation matrix, returning only unique values (lower triangle).
    Passes kwargs to pandas.DataFrame.corr method.
    """
    corr_df = df.corr(**kwargs)
    lower_triangle = pd.DataFrame(np.tril(np.ones(corr_df.shape), -1)).astype(bool)
    lower_triangle.index, lower_triangle.columns = corr_df.index, corr_df.columns
    return corr_df.where(lower_triangle)


def calculate_correlations(x, y, method="pearson"):
    """
    Calculates a Spearman (nonparametric) or a Pearson (parametric) correlation coefficient and p-value to test for non-correlation.

    :param numpy.ndarray x: array 1
    :param numpy.ndarray y: array 2
    :param str method: chooses which kind of correlation method to run
    :return: Tuple with two floats, correlation coefficient and two-tailed p-value.

    Example::

        result = calculate_correlations(x, y, method='pearson')
    """
    if method == "pearson":
        coefficient, pvalue = stats.pearsonr(x, y)
    elif method == "spearman":
        coefficient, pvalue = stats.spearmanr(x, y)

    return (coefficient, pvalue)


def run_correlation(
    df,
    alpha=0.05,
    subject="subject",
    group="group",
    method="pearson",
    correction="fdr_bh",
):
    """
    This function calculates pairwise correlations for columns in dataframe, and returns it in the shape of a edge list with 'weight' as correlation score, and the ajusted p-values.

    :param df: pandas dataframe with samples as rows and features as columns.
    :param str subject: name of column containing subject identifiers.
    :param str group: name of column containing group identifiers.
    :param str method: method to use for correlation calculation ('pearson', 'spearman').
    :param float alpha: error rate. Values velow alpha are considered significant.
    :param str correction: type of correction see apply_pvalue_correction for methods
    :return: Pandas dataframe with columns: 'node1', 'node2', 'weight', 'padj' and 'rejected'.

    Example::

        result = run_correlation(df, alpha=0.05, subject='subject', group='group', method='pearson', correction='fdr_bh')
    """
    correlation = pd.DataFrame()
    # ToDo
    # The Repeated measurements correlation calculation is too time consuming so it only runs if
    # the number of features is less than 200
    if utils.check_is_paired(df, subject, group):
        if len(df[subject].unique()) > 2:
            if len(df.columns) < 200:
                correlation = run_rm_correlation(
                    df, alpha=alpha, subject=subject, correction=correction
                )
    else:
        df = df.dropna(axis=1)._get_numeric_data()
        if not df.empty:
            r, p = run_efficient_correlation(df, method=method)
            rdf = pd.DataFrame(r, index=df.columns, columns=df.columns)
            pdf = pd.DataFrame(p, index=df.columns, columns=df.columns)
            correlation = utils.convertToEdgeList(rdf, ["node1", "node2", "weight"])
            pvalues = utils.convertToEdgeList(pdf, ["node1", "node2", "pvalue"])
            correlation = pd.merge(correlation, pvalues, on=["node1", "node2"])

            rejected, padj = apply_pvalue_correction(
                correlation["pvalue"].tolist(), alpha=alpha, method=correction
            )
            correlation["padj"] = padj
            correlation["rejected"] = rejected
            correlation = correlation[correlation.rejected]
            correlation["pvalue"] = correlation["pvalue"].apply(
                lambda x: str(round(x, 5))
            )
            correlation["padj"] = correlation["padj"].apply(lambda x: str(round(x, 5)))

    return correlation


def run_multi_correlation(
    df_dict,
    alpha=0.05,
    subject="subject",
    on=["subject", "biological_sample"],
    group="group",
    method="pearson",
    correction="fdr_bh",
):
    """
    This function merges all input dataframes and calculates pairwise correlations for all columns.

    :param dict df_dict: dictionary of pandas dataframes with samples as rows and features as columns.
    :param str subject: name of the column containing subject identifiers.
    :param str group: name of the column containing group identifiers.
    :param list on: column names to join dataframes on (must be found in all dataframes).
    :param str method: method to use for correlation calculation ('pearson', 'spearman').
    :param float alpha: error rate. Values velow alpha are considered significant.
    :param str correction: type of correction see apply_pvalue_correction for methods
    :return: Pandas dataframe with columns: 'node1', 'node2', 'weight', 'padj' and 'rejected'.

    Example::

        result = run_multi_correlation(df_dict, alpha=0.05, subject='subject', on=['subject', 'biological_sample'] , group='group', method='pearson', correction='fdr_bh')
    """
    multidf = pd.DataFrame()
    correlation = None
    for dtype in df_dict:
        if multidf.empty:
            if isinstance(df_dict[dtype], pd.DataFrame):
                multidf = df_dict[dtype]
        else:
            if isinstance(df_dict[dtype], pd.DataFrame):
                multidf = pd.merge(multidf, df_dict[dtype], how="inner", on=on)
    if not multidf.empty:
        correlation = run_correlation(
            multidf,
            alpha=alpha,
            subject=subject,
            group=group,
            method=method,
            correction=correction,
        )

    return correlation


def calculate_rm_correlation(df, x, y, subject):
    """
    Computes correlation and p-values between two columns a and b in df.

    :param df: pandas dataframe with subjects as rows and two features and columns.
    :param str x: feature a name.
    :param str y: feature b name.
    :param subject: column name containing the covariate variable.
    :return: Tuple with values for: feature a, feature b, correlation, p-value and degrees of freedom.

    Example::

        result = calculate_rm_correlation(df, x='feature a', y='feature b', subject='subject')
    """
    result = pg.rm_corr(data=df, x=x, y=y, subject=subject)

    return (
        x,
        y,
        result["r"].values[0],
        result["pval"].values[0],
        result["dof"].values[0],
    )


def run_rm_correlation(df, alpha=0.05, subject="subject", correction="fdr_bh"):
    """
    Computes pairwise repeated measurements correlations for all columns in dataframe, and returns results as an edge list with 'weight' as correlation score, p-values, degrees of freedom and ajusted p-values.

    :param df: pandas dataframe with samples as rows and features as columns.
    :param str subject: name of column containing subject identifiers.
    :param float alpha: error rate. Values velow alpha are considered significant.
    :param str correction: type of correction type see apply_pvalue_correction for methods
    :return: Pandas dataframe with columns: 'node1', 'node2', 'weight', 'pvalue', 'dof', 'padj' and 'rejected'.

    Example::

        result = run_rm_correlation(df, alpha=0.05, subject='subject', correction='fdr_bh')
    """
    rows = []
    if not df.empty:
        df = df.set_index(subject)._get_numeric_data().dropna(axis=1)
        df.columns = df.columns.astype(str)
        combinations = itertools.combinations(df.columns, 2)
        df = df.reset_index()
        for x, y in combinations:
            row = [x, y]
            subset = df[[x, y, subject]]
            row.extend(pg.rm_corr(subset, x, y, subject).values.tolist()[0])
            rows.append(row)

        correlation = pd.DataFrame(
            rows,
            columns=["node1", "node2", "weight", "dof", "pvalue", "CI95%", "power"],
        )
        rejected, padj = apply_pvalue_correction(
            correlation["pvalue"].tolist(), alpha=alpha, method=correction
        )
        correlation["padj"] = padj
        correlation["rejected"] = rejected
        correlation = correlation[correlation.rejected]
        correlation["padj"] = correlation["padj"].apply(lambda x: str(round(x, 5)))

    return correlation


def run_efficient_correlation(data, method="pearson"):
    """
    Calculates pairwise correlations and returns lower triangle of the matrix with correlation values and p-values.

    :param data: pandas dataframe with samples as index and features as columns (numeric data only).
    :param str method: method to use for correlation calculation ('pearson', 'spearman').
    :return: Two numpy arrays: correlation and p-values.

    Example::

        result = run_efficient_correlation(data, method='pearson')
    """
    matrix = data.values
    if method == "pearson":
        r = np.corrcoef(matrix, rowvar=False)
    elif method == "spearman":
        r, p = stats.spearmanr(matrix, axis=0)

    diagonal = np.triu_indices(r.shape[0], 1)
    rf = r[diagonal]
    df = matrix.shape[1] - 2
    ts = rf * rf * (df / (1 - rf * rf))
    pf = betainc(0.5 * df, 0.5, df / (df + ts))
    p = np.zeros(shape=r.shape)
    p[np.triu_indices(p.shape[0], 1)] = pf
    p[np.tril_indices(p.shape[0], -1)] = pf
    p[np.diag_indices(p.shape[0])] = np.ones(p.shape[0])

    r[diagonal] = np.nan
    p[diagonal] = np.nan

    return r, p



"""Download files from the internet."""

import requests


def download_file(url: str, local_filename: str) -> None:
    """Download a file from the internet."""
    with requests.get(url, stream=True, timeout=60) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)



import math

import numpy as np
import pandas as pd
from sklearn.impute import KNNImputer


def imputation_KNN(
    data,
    drop_cols=["group", "sample", "subject"],
    group="group",
    cutoff=0.6,
    alone=True,
):
    """
    k-Nearest Neighbors imputation for pandas dataframes with missing data. For more information visit https://github.com/iskandr/fancyimpute/blob/master/fancyimpute/knn.py.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str group: column label containing group identifiers.
    :param list drop_cols: column labels to be dropped. Final dataframe should only have gene/protein/etc identifiers as columns.
    :param float cutoff: minimum ratio of missing/valid values required to impute in each column.
    :param bool alone: if True removes all columns with any missing values.
    :return: Pandas dataframe with samples as rows and protein identifiers as columns.

    Example::

        result = imputation_KNN(data, drop_cols=['group', 'sample', 'subject'], group='group', cutoff=0.6, alone=True)
    """
    np.random.seed(112736)
    df = data.copy()
    cols = df.columns
    df = df._get_numeric_data()
    if group in data.columns:
        df[group] = data[group]
        cols = list(set(cols).difference(df.columns))
        value_cols = [c for c in df.columns if c not in drop_cols]
        for g in df[group].unique():
            missDf = df.loc[df[group] == g, value_cols]
            missDf = missDf.loc[:, missDf.notnull().mean() >= cutoff].dropna(
                axis=1, how="all"
            )
            if missDf.isnull().values.any():
                X = np.array(missDf.values, dtype=np.float64)
                X_trans = KNNImputer(n_neighbors=3).fit_transform(X)
                missingdata_df = missDf.columns.tolist()
                dfm = pd.DataFrame(
                    X_trans, index=list(missDf.index), columns=missingdata_df
                )
                df.update(dfm)
        if alone:
            df = df.dropna(axis=1)

        df = df.join(data[cols])

    return df


def imputation_mixed_norm_KNN(
    data,
    index_cols=["group", "sample", "subject"],
    shift=1.8,
    nstd=0.3,
    group="group",
    cutoff=0.6,
):
    """
    Missing values are replaced in two steps: 1) using k-Nearest Neighbors we impute protein columns with a higher ratio of missing/valid values than the defined cutoff, \
    2) the remaining missing values are replaced by random numbers that are drawn from a normal distribution.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str group: column label containing group identifiers.
    :param list index_cols: list of column labels to be set as dataframe index.
    :param float shift: specifies the amount by which the distribution used for the random numbers is shifted downwards. This is in units of the \
                        standard deviation of the valid data.
    :param float nstd: defines the width of the Gaussian distribution relative to the standard deviation of measured values. \
                        A value of 0.5 would mean that the width of the distribution used for drawing random numbers is half of the standard deviation of the data.
    :param float cutoff: minimum ratio of missing/valid values required to impute in each column.
    :return: Pandas dataframe with samples as rows and protein identifiers as columns.

    Example::

        result = imputation_mixed_norm_KNN(data, index_cols=['group', 'sample', 'subject'], shift = 1.8, nstd = 0.3, group='group', cutoff=0.6)
    """
    df = imputation_KNN(
        data, drop_cols=index_cols, group=group, cutoff=cutoff, alone=False
    )
    df = imputation_normal_distribution(
        df, index_cols=index_cols, shift=shift, nstd=nstd
    )

    return df


def imputation_normal_distribution(
    data, index_cols=["group", "sample", "subject"], shift=1.8, nstd=0.3
):
    """
    Missing values will be replaced by random numbers that are drawn from a normal distribution. The imputation is done for each sample (across all proteins) separately.
    For more information visit http://www.coxdocs.org/doku.php?id=perseus:user:activities:matrixprocessing:imputation:replacemissingfromgaussian.

    :param data: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param list index_cols: list of column labels to be set as dataframe index.
    :param float shift: specifies the amount by which the distribution used for the random numbers is shifted downwards. This is in units of the standard deviation of the valid data.
    :param float nstd: defines the width of the Gaussian distribution relative to the standard deviation of measured values. A value of 0.5 would mean that the width of the distribution used for drawing random numbers is half of the standard deviation of the data.
    :return: Pandas dataframe with samples as rows and protein identifiers as columns.

    Example::

        result = imputation_normal_distribution(data, index_cols=['group', 'sample', 'subject'], shift = 1.8, nstd = 0.3)
    """
    np.random.seed(112736)
    df = data.copy()
    if index_cols is not None:
        df = df.set_index(index_cols)

    data_imputed = df.T.sort_index()
    null_columns = data_imputed.isnull().any().index.tolist()
    for c in null_columns:
        missing = data_imputed[data_imputed[c].isnull()].index.tolist()
        std = data_imputed[c].std()
        mean = data_imputed[c].mean()
        sigma = std * nstd
        mu = mean - (std * shift)
        value = 0.0
        if (
            not math.isnan(std)
            and not math.isnan(mean)
            and not math.isnan(sigma)
            and not math.isnan(mu)
        ):
            value = np.random.normal(mu, sigma, size=len(missing))
        i = 0
        for m in missing:
            if not isinstance(value, np.ndarray):
                data_imputed.loc[m, c] = value
            else:
                data_imputed.loc[m, c] = value[i]
                i += 1

    return data_imputed.T



import pandas as pd


def transform_DataFrame(X: pd.DataFrame, fct: callable) -> pd.DataFrame:
    """Set index and columns of a DataFrame after applying a callable
    which might only return a numpy array.

    Parameters
    ----------
    X : pd.DataFrame
        Original DataFrame to be transformed
    fct : callable
        Callable to be applied to every element in the DataFrame.

    Returns
    -------
    pd.DataFrame
        Transformed DataFrame
    """
    ret = fct(X)
    ret = pd.DataFrame(ret, index=X.index, columns=X.columns)
    return ret



import os

import acore

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autosectionlabel',
    'sphinx.ext.autodoc',
    'sphinx.ext.autodoc.typehints',
    'sphinx.ext.napoleon',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.autosummary',
    'sphinx.ext.todo',
    'sphinx.ext.mathjax',
    'sphinx.ext.coverage',
    'sphinx.ext.imgmath',
    'sphinx.ext.ifconfig',
    'sphinx.ext.viewcode',
    'sphinx_new_tab_link',
    'myst_nb',
]


# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = False

autosectionlabel_prefix_document = True

## Include Python objects as they appear in source files
## Default: alphabetically ('alphabetical')
autodoc_member_order = 'bysource'
## Default flags used by autodoc directives
autodoc_default_options = {'members': True,
                            'show-inheritance': True,
                            'undoc-members': True}
## Generate autodoc stubs with summaries from code
autosummary_generate = True

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
#
# source_suffix = ['.rst', '.md']
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'acore'
copyright = "2023, Multi-omics Network Analytics Group"
author = "Multi-omics Network Analytics Group"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = acore.__version__
# The full version, including alpha/beta/rc tags.
release = acore.__version__

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store',
                    "jupyter_execute",
                    "conf.py",]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# -- General configuration ---------------------------------------------------

#  https://myst-nb.readthedocs.io/en/latest/computation/execute.html
nb_execution_mode = "auto"

myst_enable_extensions = ["dollarmath", "amsmath"]

# Plolty support through require javascript library
# https://myst-nb.readthedocs.io/en/latest/render/interactive.html#plotly
html_js_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/require.js/2.3.4/require.min.js"
]

# https://myst-nb.readthedocs.io/en/latest/configuration.html
# Execution
nb_execution_raise_on_error = True
# Rendering
nb_merge_streams = True
# maximum execution time per cell in seconds
nb_execution_timeout = 120

# https://myst-nb.readthedocs.io/en/latest/authoring/custom-formats.html#write-custom-formats
# nb_custom_formats = {".py": ["jupytext.reads", {"fmt": "py:percent"}]}
# ! better use jupytext directly to sync the notebook formats -> allows to download ipynb
# ! and see start in colab button


# Intersphinx options
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    # "scikit-learn": ("https://scikit-learn.org/stable/", None),
    # "matplotlib": ("https://matplotlib.org/stable/", None),
}

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
# See:
# https://github.com/executablebooks/MyST-NB/blob/master/docs/conf.py
# html_title = ""
html_theme = "sphinx_book_theme"
# html_logo = "_static/logo-wide.svg"
# html_favicon = "_static/logo-square.svg"
html_theme_options = {
    "github_url": "https://github.com/Multiomics-Analytics-Group/acore/",
    "repository_url": "https://github.com/Multiomics-Analytics-Group/acore/",
    "repository_branch": "main",
    "home_page_in_toc": True,
    "path_to_docs": "docs",
    "show_navbar_depth": 1,
    "use_edit_page_button": True,
    "use_repository_button": True,
    "use_download_button": True,
    "launch_buttons": {
        "colab_url": "https://colab.research.google.com"
        #     "binderhub_url": "https://mybinder.org",
        #     "notebook_interface": "jupyterlab",
    },
    "navigation_with_keys": False,
}
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']


# -- Options for HTMLHelp output ---------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'acoredoc'


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    #
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    #
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    #
    # 'preamble': '',

    # Latex figure (float) alignment
    #
    # 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'acore.tex',
     'Analytics Core Documentation',
     'Multi-omics Network Analytics Group', 'manual'),
]


# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'acore',
     'Analytics Core Documentation',
     [author], 1)
]


# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'acore',
     'Analytics Core Documentation',
     author,
     'acore',
     'Oneline description of project.',
     'Miscellaneous'),
]


# -- Setup for sphinx-apidoc -------------------------------------------------

# sphinx-apidoc needs to be called manually if Sphinx is running there.
# https://github.com/readthedocs/readthedocs.org/issues/1139

if os.environ.get("READTHEDOCS") == "True":
    from pathlib import Path

    PROJECT_ROOT = Path(__file__).parent.parent
    PACKAGE_ROOT = PROJECT_ROOT / "src" / "acore"

    def run_apidoc(_):
        from sphinx.ext import apidoc

        apidoc.main(
            [
                "--force",
                "--implicit-namespaces",
                "--module-first",
                "--separate",
                "-o",
                str(PROJECT_ROOT / "docs" / "reference"),
                str(PACKAGE_ROOT),
                str(PACKAGE_ROOT / "*.c"),
                str(PACKAGE_ROOT / "*.so"),
            ]
        )

    def setup(app):
        app.connect("builder-inited", run_apidoc)



"""Enrichment Analysis Module. Contains different functions to perform enrichment 
analysis.

Most things in this module are covered in https://www.youtube.com/watch?v=2NC1QOXmc5o
by Lars Juhl Jensen.
"""

from __future__ import annotations

import os
import re
import uuid

import gseapy as gp
import numpy as np
import pandas as pd
from scipy import stats

from acore.multiple_testing import apply_pvalue_correction

TYPE_COLS_MSG = """
columns: 'terms', 'identifiers', 'foreground',
    'background', foreground_pop, background_pop, 'pvalue', 'padj' and 'rejected'.
"""


def run_fisher(
    group1: list[int],
    group2: list[int],
    alternative: str = "two-sided",
) -> tuple[float, float]:
    """Run fisher's exact test on two groups using `scipy.stats.fisher_exact`_.

    .. _scipy.stats.fisher_exact: https://docs.scipy.org/doc/scipy/reference/generated/\
scipy.stats.fisher_exact.html

    Example::

        # annotated   not-annotated
        # group1      a               b
        # group2      c               d


        odds, pvalue = stats.fisher_exact(group1=[a, b],
                                          group2 =[c, d]
                        )
    """

    odds, pvalue = stats.fisher_exact([group1, group2], alternative)

    return (odds, pvalue)


def run_kolmogorov_smirnov(dist1, dist2, alternative="two-sided"):
    """
    Compute the Kolmogorov-Smirnov statistic on 2 samples.
    See `scipy.stats.ks_2samp`_

    .. _scipy.stats.ks_2samp: https://docs.scipy.org/doc/scipy/reference/generated/\
scipy.stats.ks_2samp.html


    :param list dist1: sequence of 1-D ndarray (first distribution to compare)
        drawn from a continuous distribution
    :param list dist2: sequence of 1-D ndarray (second distribution to compare)
        drawn from a continuous distribution
    :param str alternative: defines the alternative hypothesis (default is ‘two-sided’):
        * **'two-sided'**
        * **'less'**
        * **'greater'**
    :return: statistic float and KS statistic pvalue float Two-tailed p-value.

    Example::

        result = run_kolmogorov_smirnov(dist1, dist2, alternative='two-sided')

    """

    result = stats.ks_2samp(dist1, dist2, alternative=alternative, mode="auto")

    return result


# ! undocumented for now (find usage example)
def run_site_regulation_enrichment(
    regulation_data: pd.DataFrame,
    annotation: pd.DataFrame,
    identifier: str = "identifier",
    groups: list[str] = ("group1", "group2"),
    annotation_col: str = "annotation",
    rejected_col: str = "rejected",
    group_col: str = "group",
    method: str = "fisher",
    regex: str = "(\\w+~.+)_\\w\\d+\\-\\w+",
    correction: str = "fdr_bh",
    remove_duplicates: bool = False,
):
    r"""
    This function runs a simple enrichment analysis for significantly
    regulated protein sites in a dataset.

    :param regulation_data: pandas.DataFrame resulting from differential
        regulation analysis.
    :param annotation: pandas.DataFrame with annotations for features
        (columns: 'annotation', 'identifier' (feature identifiers), and 'source').
    :param str identifier: name of the column from annotation containing
        feature identifiers.
    :param list groups: column names from regulation_data containing
        group identifiers.
    :param str annotation_col: name of the column from annotation containing
        annotation terms.
    :param str rejected_col: name of the column from regulation_data containing
        boolean for rejected null hypothesis.
    :param str group_col: column name for new column in annotation dataframe
        determining if feature belongs to foreground or background.
    :param str method: method used to compute enrichment
        (only 'fisher' is supported currently).
    :param str regex: how to extract the annotated identifier from the site identifier
    :return: pandas.DataFrame with columns: 'terms', 'identifiers', 'foreground',
        'background', foreground_pop, background_pop, 'pvalue', 'padj' and 'rejected'.

    :raises ValueError: if regulation_data is `None` or empty.

    Example::

        result = run_site_regulation_enrichment(regulation_data,
            annotation,
            identifier='identifier',
            groups=['group1', 'group2'],
            annotation_col='annotation',
            rejected_col='rejected',
            group_col='group',
            method='fisher',
            match="(\\w+~.+)_\\w\\d+\\-\\w+"
        )
    """
    result = pd.DataFrame()
    if regulation_data is None or regulation_data.empty:
        raise ValueError("regulation_data is empty")

    new_ids = []
    # find any identifiers with a PTM and save only prot+gene identifer
    for ident in regulation_data[identifier].tolist():
        match = re.search(regex, ident)
        if match is not None:
            new_ids.append(
                match.group(1)
            )  # removes the PTM extension of the identifier of CKG
        else:
            new_ids.append(ident)
    # so this is normalizing the identifiers to ignore the PTM extension
    regulation_data[identifier] = new_ids  # matches are used as identifiers
    if remove_duplicates:
        regulation_data = regulation_data.drop_duplicates(subset=[identifier])
    result = run_regulation_enrichment(
        regulation_data,
        annotation,
        identifier,
        groups,
        annotation_col,
        rejected_col,
        group_col,
        method,
        correction,
    )

    return result


def run_up_down_regulation_enrichment(
    regulation_data: pd.DataFrame,
    annotation: pd.DataFrame,
    identifier: str = "identifier",
    groups: list[str] = ("group1", "group2"),
    annotation_col: str = "annotation",
    # rejected_col:str="rejected",
    group_col: str = "group",
    method: str = "fisher",
    min_detected_in_set: int = 2,
    correction: str = "fdr_bh",
    correction_alpha: float = 0.05,
    lfc_cutoff: float = 1,
) -> pd.DataFrame:
    """
    This function runs a simple enrichment analysis for significantly regulated proteins
    distinguishing between up- and down-regulated.

    :param pandas.DataFrame regulation_data: pandas.DataFrame resulting from differential regulation
        analysis (CKG's regulation table).
    :param pandas.DataFrame annotation: pandas.DataFrame with annotations for features
        (columns: 'annotation', 'identifier' (feature identifiers), and 'source').
    :param str identifier: name of the column from annotation containing feature identifiers.
    :param list[str] groups: column names from regulation_data containing group identifiers.
            See `pandas.DataFrame.groupby`_ for more information.
            
            .. _pandas.DataFrame.groupby: https://pandas.pydata.org/pandas-docs/stable/\
reference/api/pandas.DataFrame.groupby.html
    :param str annotation_col: name of the column from annotation containing annotation terms.
    :param str rejected_col: name of the column from regulation_data containing boolean for
        rejected null hypothesis.
    :param str group_col: column name for new column in annotation dataframe determining
        if feature belongs to foreground or background.
    :param str method: method used to compute enrichment
        (only 'fisher' is supported currently).
    :param str correction: method to be used for multiple-testing correction
    :param float alpha: adjusted p-value cutoff to define significance
    :param float lfc_cutoff: log fold-change cutoff to define practical significance
    :return: pandas.DataFrame with columns: 'terms', 'identifiers', 'foreground',
        'background', 'pvalue', 'padj', 'rejected', 'direction' and 'comparison'.

    Example::

        result = run_up_down_regulation_enrichment(
            regulation_data,
            annotation,
            identifier='identifier',
            groups=['group1',
            'group2'],
            annotation_col='annotation',
            rejected_col='rejected',
            group_col='group',
            method='fisher',
            correction='fdr_bh',
            alpha=0.05,
            lfc_cutoff=1,
        )
    """
    if isinstance(groups, str):
        groups = [groups]
    if isinstance(groups, tuple):
        groups = list(groups)
    if len(groups) != 2:
        raise ValueError("groups should contains exactly two columns.")

    ret = list()
    # In case of multiple comparisons this is used to get all possible combinations
    for g1, g2 in regulation_data.groupby(groups).groups:

        df = regulation_data.groupby(groups).get_group((g1, g2))

        padj_name = "padj"
        if "posthoc padj" in df:
            padj_name = "posthoc padj"

        df["up_pairwise_regulation"] = (df[padj_name] <= correction_alpha) & (
            df["log2FC"] >= lfc_cutoff
        )
        df["down_pairwise_regulation"] = (df[padj_name] <= correction_alpha) & (
            df["log2FC"] <= -lfc_cutoff
        )
        comparison_tag = g1 + "~" + g2

        for rej_col, direction in zip(
            ("up_pairwise_regulation", "down_pairwise_regulation"),
            ("upregulated", "downregulated"),
        ):
            _enrichment = run_regulation_enrichment(
                df,
                annotation,
                identifier=identifier,
                annotation_col=annotation_col,
                rejected_col=rej_col,
                group_col=group_col,
                method=method,
                min_detected_in_set=min_detected_in_set,
                correction=correction,
                correction_alpha=correction_alpha,
            )
            _enrichment["direction"] = direction
            _enrichment["comparison"] = comparison_tag
            ret.append(_enrichment)

    ret = pd.concat(ret)

    return ret


# ! to move
def _annotate_features(
    features: pd.Series,
    in_foreground: set[str] | list[str],
    in_background: set[str] | list[str],
) -> pd.Series:
    """
    Annotate features as foreground or background based on their presence in the
    foreground and background lists.

    :param features: pandas.Series with features and their annotations.
    :param in_foreground: list of features identifiers in the foreground.
    :type in_foreground: set or list-like
    :param in_background: list of features identifiers in the background.
    :type in_background: set or list-like
    :return: pandas.Series containing 'foreground' or 'background'.
             missing values are preserved.

    Example::

        result = _annotate_features(features, in_foreground, in_background)
    """
    in_either_or = features.isin(in_foreground) | features.isin(in_background)
    res = (
        features.where(in_either_or, np.nan)
        .mask(features.isin(in_foreground), "foreground")
        .mask(features.isin(in_background), "background")
    )
    return res


def run_regulation_enrichment(
    regulation_data: pd.DataFrame,
    annotation: pd.DataFrame,
    identifier: str = "identifier",
    annotation_col: str = "annotation",
    rejected_col: str = "rejected",
    group_col: str = "group",
    method: str = "fisher",
    min_detected_in_set: int = 2,
    correction: str = "fdr_bh",
    correction_alpha: float = 0.05,
) -> pd.DataFrame:
    """
    This function runs a simple enrichment analysis for significantly regulated features
    in a dataset.

    :param regulation_data: pandas.DataFrame resulting from differential regulation analysis.
    :param annotation: pandas.DataFrame with annotations for features
        (columns: 'annotation', 'identifier' (feature identifiers), and 'source').
    :param str identifier: name of the column from annotation containing feature identifiers.
        It should also be present in `regulation_data`.
    :param str annotation_col: name of the column from annotation containing annotation terms.
    :param str rejected_col: name of the column from `regulation_data` containing boolean for
        rejected null hypothesis.
    :param str group_col: column name for new column in annotation dataframe determining
        if feature belongs to foreground or background.
    :param str method: method used to compute enrichment (only 'fisher' is supported currently).
    :param str correction: method to be used for multiple-testing correction
    :return: pandas.DataFrame with columns: 'terms', 'identifiers', 'foreground',
        'background', 'foreground_pop', 'background_pop', 'pvalue', 'padj' and 'rejected'.

    Example::

        result = run_regulation_enrichment(
            regulation_data,
            annotation,
            identifier='identifier',
            annotation_col='annotation',
            rejected_col='rejected',
            group_col='group',
            method='fisher',
            min_detected_in_set=2,
            correction='fdr_bh',
            correction_alpha=0.05,
         )
    """
    # ? can we remove NA features in that column?
    mask_rejected = regulation_data[rejected_col].astype(bool)
    foreground_list = regulation_data.loc[mask_rejected, identifier].unique()
    background_list = regulation_data.loc[~mask_rejected, identifier].unique()
    foreground_pop = len(foreground_list)
    background_pop = len(regulation_data[identifier].unique())
    # needs to allow for missing annotations
    annotation[group_col] = _annotate_features(
        features=annotation[identifier],
        in_foreground=foreground_list,
        in_background=background_list,
    )
    annotation = annotation.dropna(subset=[group_col])

    result = run_enrichment(
        annotation,
        foreground_id="foreground",
        background_id="background",
        foreground_pop=foreground_pop,
        background_pop=background_pop,
        annotation_col=annotation_col,
        group_col=group_col,
        identifier_col=identifier,
        method=method,
        correction=correction,
        min_detected_in_set=min_detected_in_set,
        correction_alpha=correction_alpha,
    )

    return result


def run_enrichment(
    data: pd.DataFrame,
    foreground_id: str,
    background_id: str,
    foreground_pop: int,
    background_pop: int,
    min_detected_in_set: int = 2,
    annotation_col: str = "annotation",
    group_col: str = "group",
    identifier_col: str = "identifier",
    method: str = "fisher",
    correction: str = "fdr_bh",
    correction_alpha: float = 0.05,
) -> pd.DataFrame:
    """
    Computes enrichment of the foreground relative to a given backgroung,
    using Fisher's exact test, and corrects for multiple hypothesis testing.

    :param data: pandas.DataFrame with annotations for dataset features
        (columns: 'annotation', 'identifier', 'group').
    :param str foreground_id: group identifier of features that belong to the foreground.
    :param str background_id: group identifier of features that belong to the background.
    :param int foreground_pop: number of features in the foreground.
    :param int background_pop: number of features in the background.
    :param str annotation_col: name of the column containing annotation terms.
    :param str group_col: name of column containing the group identifiers.
    :param str identifier_col: name of column containing dependent variables identifiers.
    :param str method: method used to compute enrichment (only 'fisher' is supported currently).
    :param str correction: method to be used for multiple-testing correction.
    :param float correction_alpha: adjusted p-value cutoff to define significance.
    :return: pandas.DataFrame with columns: annotation terms, features,
        number of foregroung/background features in each term,
        p-values and corrected p-values
        (columns: 'terms', 'identifiers', 'foreground',
        'background', 'pvalue', 'padj' and 'rejected').

    Example::

        result = run_enrichment(
            data,
            foreground='foreground',
            background='background',
            foreground_pop=len(foreground_list),
            background_pop=len(background_list),
            annotation_col='annotation',
            group_col='group',
            identifier_col='identifier',
            method='fisher',
         )
    """
    if method != "fisher":
        raise ValueError("Only Fisher's exact test is supported at the moment.")

    result = pd.DataFrame()
    terms = []
    ids = []
    pvalues = []
    fnum = []
    bnum = []
    countsdf = (
        data.groupby([annotation_col, group_col])
        .agg(["count"])[(identifier_col, "count")]
        .reset_index()
    )
    countsdf.columns = [annotation_col, group_col, "count"]
    for annotation in countsdf.loc[
        countsdf[group_col] == foreground_id, annotation_col
    ].unique():
        counts = countsdf[countsdf[annotation_col] == annotation]
        num_foreground = counts.loc[counts[group_col] == foreground_id, "count"].values
        num_background = counts.loc[counts[group_col] == background_id, "count"].values
        # ! counts should always be of length one count? squeeze?
        if len(num_foreground) == 1:
            num_foreground = num_foreground[0]
        if len(num_background) == 1:
            num_background = num_background[0]
        else:
            num_background = 0
        if num_foreground >= min_detected_in_set:
            _, pvalue = run_fisher(
                [num_foreground, foreground_pop - num_foreground],
                [num_background, background_pop - foreground_pop - num_background],
            )
            fnum.append(num_foreground)
            bnum.append(num_background)
            terms.append(annotation)
            pvalues.append(pvalue)
            ids.append(
                ",".join(
                    data.loc[
                        (data[annotation_col] == annotation)
                        & (data[group_col] == foreground_id),
                        identifier_col,
                    ]
                )
            )
    if len(pvalues) >= 1:
        rejected, padj = apply_pvalue_correction(
            pvalues,
            alpha=correction_alpha,
            method=correction,
        )
        result = pd.DataFrame(
            {
                "terms": terms,
                "identifiers": ids,
                "foreground": fnum,
                "background": bnum,
                "foreground_pop": foreground_pop,
                "background_pop": background_pop,
                "pvalue": pvalues,
                "padj": padj,
                "rejected": rejected.astype(bool),
            }
        )
        result = result.sort_values(by="padj", ascending=True)

    return result


def run_ssgsea(
    data: pd.DataFrame,
    annotation: str,
    set_index: list[str] = None,
    annotation_col: str = "annotation",
    identifier_col: str = "identifier",
    outdir: str = "tmp",
    min_size: int = 15,
    max_size: int = 500,
    scale: bool = False,
    permutations: int = 0,
) -> pd.DataFrame:
    """
    Project each sample within a data set onto a space of gene set enrichment scores using
    the single sample gene set enrichment analysis (ssGSEA) projection methodology
    described in Barbie et al., 2009:
    https://www.nature.com/articles/nature08460#Sec3 (search "Single Sample" GSEA).

    :param pd.DataFrame data: pandas.DataFrame with the quantified features (i.e. subject x proteins)
    :param str annotation: pandas.DataFrame with the annotation to be used in the enrichment
        (i.e. CKG pathway annotation file)
    :param list[str] set_index: column/s to be used as index. Enrichment will be calculated
        for these values (i.e ["subject"] will return subjects x pathways matrix of
        enrichment scores)
    :param str annotation_col: name of the column containing annotation terms.
    :param str identifier_col: name of column containing dependent variables identifiers.
    :param str out_dir: directory path where results will be stored
        (default None, tmp folder is used)
    :param int min_size: minimum number of features (i.e. proteins) in enriched terms
        (i.e. pathways)
    :param int max_size: maximum number of features (i.e. proteins) in enriched terms
        (i.e. pathways)
    :param bool scale: whether or not to scale the data
    :param int permutations: number of permutations used in the ssgsea analysis
    :return: pandas.DataFrame containing unnormalized enrichment scores (`ES`) for each sample,
        and normalized enrichment scores (`NES`) with the enriched `Term` and sample `Name`.
    :rtype: pandas.DataFrame

    Example::

        stproject = "P0000008"
        p = project.Project(
            stproject,
            datasets={},
            knowledge=None,
            report={},
            configuration_files=None,
        )
        p.build_project(False)
        p.generate_report()

        proteomics_dataset = p.get_dataset("proteomics")
        annotations = proteomics_dataset.get_dataframe("pathway annotation")
        processed = proteomics_dataset.get_dataframe('processed')

        result = run_ssgsea(
            processed,
            annotations,
            annotation_col='annotation',
            identifier_col='identifier',
            set_index=['group',
            'sample',
            'subject'],
            outdir=None,
            min_size=10,
            scale=False,
            permutations=0
        )
    """
    df = data.copy()
    if not os.path.exists(outdir):
        os.makedirs(outdir)

    # Comine columns to create a unique name for each set (?)
    name = []
    if set_index is None:
        index = data.index.to_frame()
        set_index = index.columns.tolist()
    else:
        index = data[set_index]
        df = df.drop(set_index, axis=1)

    for _, row in index.iterrows():
        name.append(
            "_".join(row[set_index].tolist())
        )  # this assumes strings as identifiers

    df["Name"] = name
    index.index = name
    df = df.set_index("Name").transpose()

    if not annotation_col in annotation:
        raise ValueError(
            f"Missing Annotation Column: {annotation_col} as specified by `annotation_col`"
        )

    if not identifier_col in annotation:
        raise ValueError(
            f"Missing Identifier Column: {identifier_col} as specified by `identifier_col`"
        )

    grouped_annotations = (
        annotation.groupby(annotation_col)[identifier_col].apply(list).reset_index()
    )
    fid = uuid.uuid4()
    file_path = os.path.join(outdir, str(fid) + ".gmt")
    with open(file_path, "w", encoding="utf8") as out:
        for _, row in grouped_annotations.iterrows():
            out.write(
                row[annotation_col]
                + "\t"
                + "\t".join(list(filter(None, row[identifier_col])))
                + "\n"
            )
    enrichment = gp.ssgsea(
        data=df,
        gene_sets=str(file_path),
        outdir=outdir,
        min_size=min_size,
        max_size=max_size,
        scale=scale,
        permutation_num=permutations,
        no_plot=True,
        processes=1,
        seed=10,
        format="png",
    )
    result = pd.DataFrame(enrichment.res2d).set_index("Name")
    # potentially return wide format in separate format
    # result = {"es": enrichment_es, "nes": enrichment_nes}
    return result



import unittest

import numpy as np
import pandas as pd
from scipy import stats

import acore.enrichment_analysis as ea


class TestRunFisher(unittest.TestCase):
    def test_run_fisher(self):
        group1 = [10, 5]
        group2 = [8, 12]
        alternative = "two-sided"

        expected_odds, expected_pvalue = stats.fisher_exact(
            [[10, 5], [8, 12]], alternative
        )

        result = ea.run_fisher(group1, group2, alternative=alternative)

        self.assertEqual(result[0], expected_odds)
        self.assertEqual(result[1], expected_pvalue)


class TestRunKolmogorovSmirnov(unittest.TestCase):
    def test_run_kolmogorov_smirnov(self):
        dist1 = [1, 2, 3, 4, 5]
        dist2 = [1, 2, 3, 4, 6]
        alternative = "two-sided"

        expected_result = stats.ks_2samp(
            dist1, dist2, alternative=alternative, mode="auto"
        )

        result = ea.run_kolmogorov_smirnov(dist1, dist2, alternative=alternative)

        self.assertEqual(result[0], expected_result.statistic)
        self.assertEqual(result[1], expected_result.pvalue)


def test__annotate_features():
    expected = pd.Series(
        [
            "foreground",
            "foreground",
            "background",
            "foreground",
            "background",
            "background",
            np.nan,
        ]
    )

    features = pd.Series(["G1", "G2", "G3", "G4", "G5", "G6", "G9"])
    in_foreground = ["G1", "G2", "G4"]
    in_background = ["G3", "G5", "G6"]
    actual = ea._annotate_features(features, in_foreground, in_background)
    pd.testing.assert_series_equal(expected, actual)


def test_run_regulation_enrichment():
    """Integration test for run_regulation_enrichment. Indirectly tests
    run_enrichment from enrichment_analysis module."""
    annotation = {
        "annotation": ["path1", "path1", "path1", "path2", "path2", "path3", "path3"],
        "identifier": ["gene1", "gene2", "gene3", "gene1", "gene5", "gene6", "gene9"],
        "source": ["GO", "GO", "GO", "GO_P", "GO_P", "GO_P", "GO_P"],
    }
    annotation = pd.DataFrame(annotation)
    regulation_res = {
        "identifier": ["gene1", "gene2", "gene3", "gene4", "gene5", "gene6"],
        "rejected": [True, True, False, False, True, True],
    }
    regulation_res = pd.DataFrame(regulation_res)

    actual = ea.run_regulation_enrichment(
        regulation_data=regulation_res,
        annotation=annotation,
        min_detected_in_set=1,
    )

    expected = pd.DataFrame(
        {
            "terms": ["path1", "path2", "path3"],
            "identifiers": ["gene1,gene2", "gene1,gene5", "gene6"],
            "foreground": [2, 2, 1],
            "background": [1, 0, 0],
            "foreground_pop": [4, 4, 4],
            "background_pop": [6, 6, 6],
            "pvalue": [1.0, 0.4666666666666667, 1.0],
            "padj": [1.0, 1.0, 1.0],
            "rejected": [False, False, False],
        }
    )
    assert expected.equals(actual)


if __name__ == "__main__":
    unittest.main()



"""Strategies for data normalization used by `normalize_data`."""

import pandas as pd
from sklearn import preprocessing


# ! update docstring: example is not correct
def median_zero_normalization(data, normalize="samples"):
    """
    This function normalizes each sample by using its median.

    :param data:
    :param str normalize: whether the normalization should be done by 'features' (columns)
                          or 'samples' (rows)
    :return: pandas.DataFrame.

    Example::
        data = pd.DataFrame({'a': [2,5,4,3,3], 'b':[4,4,6,5,3], 'c':[4,14,8,8,9]})
        result = median_normalization(data, normalize='samples')
        result
                a         b         c
            0 -1.333333  0.666667  0.666667
            1 -2.666667 -3.666667  6.333333
            2 -2.000000  0.000000  2.000000
            3 -2.333333 -0.333333  2.666667
            4 -2.000000 -2.000000  4.000000
    """
    if normalize is None or normalize == "samples":
        norm_data = data.sub(data.median(axis=1), axis=0)
    else:
        norm_data = data.sub(data.median(axis=0), axis=1)

    return norm_data


# ! Update docstring: example is not correct
def median_normalization(data, normalize="samples"):
    """
    This function normalizes each sample by using its median.

    :param data:
    :param str normalize: whether the normalization should be done by 'features' (columns)
                          or 'samples' (rows)
    :return: pandas.DataFrame.

    Example::
        data = pd.DataFrame({'a': [2,5,4,3,3], 'b':[4,4,6,5,3], 'c':[4,14,8,8,9]})
        result = median_normalization(data, normalize='samples')
        result
                a         b         c
            0 -1.333333  0.666667  0.666667
            1 -2.666667 -3.666667  6.333333
            2 -2.000000  0.000000  2.000000
            3 -2.333333 -0.333333  2.666667
            4 -2.000000 -2.000000  4.000000
    """
    if normalize is None or normalize == "samples":
        norm_data = data.sub(data.median(axis=1) - data.median(axis=1).median(), axis=0)
    else:
        norm_data = data.sub(data.median(axis=0) - data.median(axis=0).median(), axis=1)

    return norm_data


def zscore_normalization(data, normalize="samples"):
    """
    This function normalizes each sample by using its mean and standard deviation
    (mean=0, std=1).

    :param data:
    :param str normalize: whether the normalization should be done by 'features' (columns)
                          or 'samples' (rows)
    :return: pandas.DataFrame.

    Example::
        data = pd.DataFrame({'a': [2,5,4,3,3], 'b':[4,4,6,5,3], 'c':[4,14,8,8,9]})
        result = zscore_normalization(data, normalize='samples')
        result
                  a         b         c
                0 -1.154701  0.577350  0.577350
                1 -0.484182 -0.665750  1.149932
                2 -1.000000  0.000000  1.000000
                3 -0.927173 -0.132453  1.059626
                4 -0.577350 -0.577350  1.154701
    """
    if normalize is None or normalize == "samples":
        norm_data = data.sub(data.mean(axis=1), axis=0).div(data.std(axis=1), axis=0)

    else:
        norm_data = data.sub(data.mean(axis=0), axis=1).div(data.std(axis=0), axis=1)

    return norm_data


def median_polish_normalization(data, max_iter=250):
    """
    This function iteratively normalizes each sample and each feature to its
    median until medians converge.

    :param data:
    :param int max_iter: number of maximum iterations to prevent infinite loop.
    :return: pandas.DataFrame.

    Example::
        data = pd.DataFrame({'a': [2,5,4,3,3], 'b':[4,4,6,5,3], 'c':[4,14,8,8,9]})
        result = median_polish_normalization(data, max_iter = 10)
        result
                a    b     c
            0  2.0  4.0   7.0
            1  5.0  7.0  10.0
            2  4.0  6.0   9.0
            3  3.0  5.0   8.0
            4  3.0  5.0   8.0
    """
    mediandf = data.copy()
    for _ in range(max_iter):
        row_median = mediandf.median(axis=1)
        mediandf = mediandf.sub(row_median, axis=0)
        col_median = mediandf.median(axis=0)
        mediandf = mediandf.sub(col_median, axis=1)
        if (mediandf.median(axis=0) == 0).all() and (
            mediandf.median(axis=1) == 0
        ).all():
            break

    norm_data = data - mediandf

    return norm_data


def quantile_normalization(data) -> pd.DataFrame:
    """
    Applies quantile normalization to each column in pandas.DataFrame.

    :param data: pandas.DataFrame with features as columns and samples as rows.
    :return: pandas.DataFrame

    Example::
        data = pd.DataFrame({'a': [2,5,4,3,3], 'b':[4,4,6,5,3], 'c':[4,14,8,8,9]})
        result = quantile_normalization(data)
        result
                a    b    c
            0  3.2  4.6  4.6
            1  4.6  3.2  8.6
            2  3.2  4.6  8.6
            3  3.2  4.6  8.6
            4  3.2  3.2  8.6
    """
    rank_mean = (
        data.T.stack().groupby(data.T.rank(method="first").stack().astype(int)).mean()
    )
    normdf = data.T.rank(method="min").stack().astype(int).map(rank_mean).unstack().T

    return normdf


def linear_normalization(data, method="l1", normalize="samples") -> pd.DataFrame:
    """
    This function scales input data to a unit norm. For more information visit:
    https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.normalize.html

    :param data: pandas.DataFrame with samples as rows and features as columns.
    :param str method: norm to use to normalize each non-zero sample or non-zero feature
                        (depends on axis).
    :param str normalize: axis used to normalize the data along. If 'samples',
                independently normalize each sample, if 'features' normalize each feature.
    :return: pandas.DataFrame

    Example::
        data = pd.DataFrame({'a': [2,5,4,3,3], 'b':[4,4,6,5,3], 'c':[4,14,8,8,9]})
        result = linear_normalization(data, method = "l1", by = 'samples')
        result
                a         b         c
            0  0.117647  0.181818  0.093023
            1  0.294118  0.181818  0.325581
            2  0.235294  0.272727  0.186047
            3  0.176471  0.227273  0.186047
            4  0.176471  0.136364  0.209302
    """
    if normalize is None or normalize == "samples":
        normvalues = preprocessing.normalize(
            data.fillna(0).values, norm=method, axis=0, copy=True, return_norm=False
        )
    else:
        normvalues = preprocessing.normalize(
            data.fillna(0).values, norm=method, axis=1, copy=True, return_norm=False
        )

    normdf = pd.DataFrame(normvalues, index=data.index, columns=data.columns)

    return normdf



import kmapper as km
import numpy as np
from sklearn import cluster, ensemble


def run_mapper(
    data,
    lenses=["l2norm"],
    n_cubes=15,
    overlap=0.5,
    n_clusters=3,
    linkage="complete",
    affinity="correlation",
):
    """

    :param data:
    :param lenses:
    :param n_cubes:
    :param overlap:
    :param n_clusters:
    :param linkage:
    :param affinity:
    :return:

    """

    X = data._get_numeric_data()
    labels = {i: data.index[i] for i in range(len(data.index))}

    model = ensemble.IsolationForest(random_state=1729)
    model.fit(X)
    lens1 = model.decision_function(X).reshape((X.shape[0], 1))

    # Create another 1-D lens with L2-norm
    mapper = km.KeplerMapper(verbose=0)
    lens2 = mapper.fit_transform(X, projection=lenses[0])

    # Combine both lenses to get a 2-D [Isolation Forest, L^2-Norm] lens
    lens = np.c_[lens1, lens2]

    # Define the simplicial complex
    simplicial_complex = mapper.map(
        lens,
        X,
        clusterer=cluster.AgglomerativeClustering(
            n_clusters=n_clusters, linkage=linkage, affinity=affinity
        ),
    )

    return simplicial_complex, {"labels": labels}



import re

import numpy as np
import pandas as pd
import pingouin as pg
import scipy.stats
import statsmodels.api as sm
from statsmodels.formula.api import ols

import acore.utils
from acore.multiple_testing import (
    apply_pvalue_correction,
    apply_pvalue_permutation_fdrcorrection,
    correct_pairwise_ttest,
    get_max_permutations,
)


# njab.stats.groups_comparision.py (partly renamed functions)
def calc_means_between_groups(
    df: pd.DataFrame,
    boolean_array: pd.Series,
    event_names: tuple[str, str] = ("1", "0"),
) -> pd.DataFrame:
    """Mean comparison between groups"""
    sub = df.loc[boolean_array].describe().iloc[:3]
    sub["event"] = event_names[0]
    sub = sub.set_index("event", append=True).swaplevel()
    ret = sub
    sub = df.loc[~boolean_array].describe().iloc[:3]
    sub["event"] = event_names[1]
    sub = sub.set_index("event", append=True).swaplevel()
    ret = pd.concat([ret, sub])
    ret.columns.name = "variable"
    ret.index.names = ("event", "stats")
    return ret.T


def calc_ttest(
    df: pd.DataFrame, boolean_array: pd.Series, variables: list[str]
) -> pd.DataFrame:
    """Calculate t-test for each variable in `variables` between two groups defined
    by boolean array."""
    ret = []
    for var in variables:
        _ = pg.ttest(df.loc[boolean_array, var], df.loc[~boolean_array, var])
        ret.append(_)
    ret = pd.concat(ret)
    ret = ret.set_index(variables)
    ret.columns.name = "ttest"
    ret.columns = pd.MultiIndex.from_product(
        [["ttest"], ret.columns], names=("test", "var")
    )
    return ret


def run_diff_analysis(
    df: pd.DataFrame,
    boolean_array: pd.Series,
    event_names: tuple[str, str] = ("1", "0"),
    ttest_vars=("alternative", "p-val", "cohen-d"),
) -> pd.DataFrame:
    """Differential analysis procedure between two groups. Calculaes
    mean per group and t-test for each variable in `vars` between two groups."""
    ret = calc_means_between_groups(
        df, boolean_array=boolean_array, event_names=event_names
    )
    ttests = calc_ttest(df, boolean_array=boolean_array, variables=ret.index)
    ret = ret.join(ttests.loc[:, pd.IndexSlice[:, ttest_vars]])
    return ret


# end njab.stats.groups_comparision.py


def calculate_ttest(
    df,
    condition1,
    condition2,
    paired=False,
    is_logged=True,
    non_par=False,
    tail="two-sided",
    correction="auto",
    r=0.707,
):
    """
    Calculates the t-test for the means of independent samples belonging to two different
    groups. For more information visit
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_ind.html.

    :param df: pandas dataframe with groups and subjects as rows and protein identifier
               as column.
    :param str condition1: identifier of first group.
    :param str condition2: ientifier of second group.
    :param bool is_logged: data is logged transformed
    :param bool non_par: if True, normality and variance equality assumptions are checked
                         and non-parametric test Mann Whitney U test if not passed
    :return: Tuple with t-statistics, two-tailed p-value, mean of first group,
             mean of second group and logfc.

    Example::

        result = calculate_ttest(df, 'group1', 'group2')
    """
    t = None
    pvalue = np.nan
    group1 = df[[condition1]].values
    group2 = df[[condition2]].values

    mean1 = group1.mean()
    std1 = group1.std()
    mean2 = group2.mean()
    std2 = group2.std()
    if is_logged:
        fc = mean1 - mean2
    else:
        fc = mean1 / mean2

    test = "t-Test"
    if not non_par:
        result = pg.ttest(
            df[condition1],
            df[condition2],
            paired=paired,
            alternative=tail,
            correction=correction,
            r=r,
        )
    else:
        test = "Mann Whitney"
        result = pg.mwu(group1, group2, alternative=tail)

    if "T" in result.columns:
        t = result["T"].values[0]
    elif "U-val" in result.columns:
        t = result["U-val"].values[0]
    if "p-val" in result.columns:
        pvalue = result["p-val"].values[0]

    return (t, pvalue, mean1, mean2, std1, std2, fc, test)


def calculate_THSD(df, column, group="group", alpha=0.05, is_logged=True):
    """
    Pairwise Tukey-HSD posthoc test using pingouin stats.
    For more information visit https://pingouin-stats.org/generated/pingouin.pairwise_tukey.html

    :param df: pandas dataframe with group and protein identifier as columns
    :param str column: column containing the protein identifier
    :param str group: column label containing the between factor
    :param float alpha: significance level
    :return: Pandas dataframe.

    Example::

        result = calculate_THSD(df, column='HBG2~P69892', group='group', alpha=0.05)
    """
    posthoc = None
    posthoc = pg.pairwise_tukey(data=df, dv=column, between=group)
    posthoc.columns = [
        "group1",
        "group2",
        "mean(group1)",
        "mean(group2)",
        "log2FC",
        "std_error",
        "t-statistics",
        "posthoc pvalue",
        "effsize",
    ]
    posthoc["efftype"] = "hedges"
    posthoc = complement_posthoc(posthoc, identifier=column, is_logged=is_logged)

    return posthoc


def calculate_pairwise_ttest(
    df, column, subject="subject", group="group", correction="none", is_logged=True
):
    """
    Performs pairwise t-test using pingouin, as a posthoc test, and calculates fold-changes. For more information visit https://pingouin-stats.org/generated/pingouin.pairwise_ttests.html.

    :param df: pandas dataframe with subject and group as rows and protein identifier as column.
    :param str column: column label containing the dependant variable
    :param str subject: column label containing subject identifiers
    :param str group: column label containing the between factor
    :param str correction: method used for testing and adjustment of p-values.
    :return: Pandas dataframe with means, standard deviations, test-statistics, degrees of freedom and effect size columns.

    Example::

        result = calculate_pairwise_ttest(df, 'protein a', subject='subject', group='group', correction='none')
    """

    posthoc_columns = [
        "Contrast",
        "group1",
        "group2",
        "mean(group1)",
        "std(group1)",
        "mean(group2)",
        "std(group2)",
        "posthoc Paired",
        "posthoc Parametric",
        "posthoc T-Statistics",
        "posthoc dof",
        "posthoc tail",
        "posthoc pvalue",
        "posthoc BF10",
        "posthoc effsize",
    ]
    valid_cols = [
        "group1",
        "group2",
        "mean(group1)",
        "std(group1)",
        "mean(group2)",
        "std(group2)",
        "posthoc Paired",
        "posthoc Parametric",
        "posthoc T-Statistics",
        "posthoc dof",
        "posthoc tail",
        "posthoc pvalue",
        "posthoc BF10",
        "posthoc effsize",
    ]
    posthoc = df.pairwise_ttests(
        dv=column,
        between=group,
        subject=subject,
        effsize="hedges",
        return_desc=True,
        padjust=correction,
    )
    posthoc.columns = posthoc_columns
    posthoc = posthoc[valid_cols]
    posthoc = complement_posthoc(posthoc, column, is_logged)
    posthoc["efftype"] = "hedges"

    return posthoc


def complement_posthoc(posthoc, identifier, is_logged):
    """
    Calculates fold-changes after posthoc test.

    :param posthoc: pandas dataframe from posthoc test. Should have at least columns 'mean(group1)' and 'mean(group2)'.
    :param str identifier: feature identifier.
    :return: Pandas dataframe with additional columns 'identifier', 'log2FC' and 'FC'.
    """
    posthoc["identifier"] = identifier
    if is_logged:
        posthoc["log2FC"] = posthoc["mean(group1)"] - posthoc["mean(group2)"]
        posthoc["FC"] = posthoc["log2FC"].apply(lambda x: np.power(2, x))
    else:
        posthoc["FC"] = posthoc["mean(group1)"] / posthoc["mean(group2)"]

    return posthoc


def calculate_anova(df, column, group="group"):
    """
    Calculates one-way ANOVA using pingouin.

    :param df: pandas dataframe with group as rows and protein identifier as column
    :param str column: name of the column in df to run ANOVA on
    :param str group: column with group identifiers
    :return: Tuple with t-statistics and p-value.
    """
    aov_result = pg.anova(data=df, dv=column, between=group)
    df1, df2, t, pvalue = aov_result[["ddof1", "ddof2", "F", "p-unc"]].values.tolist()[
        0
    ]

    return (column, df1, df2, t, pvalue)


def calculate_ancova(data, column, group="group", covariates=[]):
    """
    Calculates one-way ANCOVA using pingouin.

    :param df: pandas dataframe with group as rows and protein identifier as column
    :param str column: name of the column in df to run ANOVA on
    :param str group: column with group identifiers
    :param list covariates: list of covariates (columns in df)
    :return: Tuple with column, F-statistics and p-value.
    """
    ancova_result = pg.ancova(data=data, dv=column, between=group, covar=covariates)
    t, df, pvalue = (
        ancova_result.loc[ancova_result["Source"] == group, ["F", "DF", "p-unc"]]
        .values.tolist()
        .pop()
    )

    return (column, df, df, t, pvalue)


def calculate_repeated_measures_anova(df, column, subject="subject", within="group"):
    """
    One-way and two-way repeated measures ANOVA using pingouin stats.

    :param df: pandas dataframe with samples as rows and protein identifier as column. Data must be in long-format for two-way repeated measures.
    :param str column: column label containing the dependant variable
    :param str subject: column label containing subject identifiers
    :param str within: column label containing the within factor
    :return: Tuple with protein identifier, t-statistics and p-value.

    Example::

        result = calculate_repeated_measures_anova(df, 'protein a', subject='subject', within='group')
    """
    df1 = np.nan
    df2 = np.nan
    t = np.nan
    pvalue = np.nan
    try:
        aov_result = pg.rm_anova(
            data=df,
            dv=column,
            within=within,
            subject=subject,
            detailed=True,
            correction=True,
        )
        t, pvalue = aov_result.loc[0, ["F", "p-unc"]].values.tolist()
        df1, df2 = aov_result["DF"]
    except Exception as e:
        print(
            "Repeated measurements Anova for column: {} could not be calculated."
            " Error {}".format(column, e)
        )

    return (column, df1, df2, t, pvalue)


def calculate_mixed_anova(
    df, column, subject="subject", within="group", between="group2"
):
    """
    One-way and two-way repeated measures ANOVA using pingouin stats.

    :param df: pandas dataframe with samples as rows and protein identifier as column. Data must be in long-format for two-way repeated measures.
    :param str column: column label containing the dependant variable
    :param str subject: column label containing subject identifiers
    :param str within: column label containing the within factor
    :param str within: column label containing the between factor
    :return: Tuple with protein identifier, t-statistics and p-value.

    Example::

        result = calculate_mixed_anova(df, 'protein a', subject='subject', within='group', between='group2')
    """
    try:
        aov_result = pg.mixed_anova(
            data=df,
            dv=column,
            within=within,
            between=between,
            subject=subject,
            correction=True,
        )
        aov_result["identifier"] = column
    except Exception as e:
        print(
            "Mixed Anova for column: {} could not be calculated. Error {}".format(
                column, e
            )
        )

    return aov_result[["identifier", "DF1", "DF2", "F", "p-unc", "Source"]]


def run_anova(
    df,
    alpha=0.05,
    drop_cols=["sample", "subject"],
    subject="subject",
    group="group",
    permutations=0,
    correction="fdr_bh",
    is_logged=True,
    non_par=False,
):
    """
    Performs statistical test for each protein in a dataset.
    Checks what type of data is the input (paired, unpaired or repeated measurements) and performs posthoc tests for multiclass data.
    Multiple hypothesis correction uses permutation-based if permutations>0 and Benjamini/Hochberg if permutations=0.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str subject: column with subject identifiers
    :param str group: column with group identifiers
    :param list drop_cols: column labels to be dropped from the dataframe
    :param float alpha: error rate for multiple hypothesis correction
    :param int permutations: number of permutations used to estimate false discovery rates.
    :param bool non_par: if True, normality and variance equality assumptions are checked and non-parametric test Mann Whitney U test if not passed
    :return: Pandas dataframe with columns 'identifier', 'group1', 'group2', 'mean(group1)', 'mean(group2)', 'Log2FC', 'std_error', 'tail', 't-statistics', 'posthoc pvalue', 'effsize', 'efftype', 'FC', 'rejected', 'F-statistics', 'p-value', 'correction', '-log10 p-value', and 'method'.

    Example::

        result = run_anova(df, alpha=0.05, drop_cols=["sample",'subject'], subject='subject', group='group', permutations=50)
    """
    res = pd.DataFrame()
    if subject is not None and acore.utils.check_is_paired(df, subject, group):
        groups = df[group].unique()
        drop_cols = [d for d in drop_cols if d != subject]
        if len(groups) == 2:
            res = run_ttest(
                df,
                groups[0],
                groups[1],
                alpha=alpha,
                drop_cols=drop_cols,
                subject=subject,
                group=group,
                paired=True,
                correction=correction,
                permutations=permutations,
                is_logged=is_logged,
                non_par=non_par,
            )
        elif len(groups) > 2:
            res = run_repeated_measurements_anova(
                df,
                alpha=alpha,
                drop_cols=drop_cols,
                subject=subject,
                within=group,
                permutations=0,
                is_logged=is_logged,
            )
    elif len(df[group].unique()) == 2:
        groups = df[group].unique()
        drop_cols = [d for d in drop_cols if d != subject]
        res = run_ttest(
            df,
            groups[0],
            groups[1],
            alpha=alpha,
            drop_cols=drop_cols,
            subject=subject,
            group=group,
            paired=False,
            correction=correction,
            permutations=permutations,
            is_logged=is_logged,
            non_par=non_par,
        )
    elif len(df[group].unique()) > 2:
        df = df.drop(drop_cols, axis=1)
        aov_results = []
        pairwise_results = []
        for col in df.columns.drop(group).tolist():
            aov = calculate_anova(df[[group, col]], column=col, group=group)
            aov_results.append(aov)
            pairwise_result = calculate_pairwise_ttest(
                df[[group, col]],
                column=col,
                subject=subject,
                group=group,
                is_logged=is_logged,
            )
            pairwise_cols = pairwise_result.columns
            pairwise_results.extend(pairwise_result.values.tolist())
        df = df.set_index([group])
        res = format_anova_table(
            df,
            aov_results,
            pairwise_results,
            pairwise_cols,
            group,
            permutations,
            alpha,
            correction,
        )
        res["Method"] = "One-way anova"
        res = correct_pairwise_ttest(res, alpha, correction)

    return res


def run_ancova(
    df,
    covariates,
    alpha=0.05,
    drop_cols=["sample", "subject"],
    subject="subject",
    group="group",
    permutations=0,
    correction="fdr_bh",
    is_logged=True,
    non_par=False,
):
    """
    Performs statistical test for each protein in a dataset.
    Checks what type of data is the input (paired, unpaired or repeated measurements) and performs posthoc tests for multiclass data.
    Multiple hypothesis correction uses permutation-based if permutations>0 and Benjamini/Hochberg if permutations=0.

    :param df: pandas dataframe with samples as rows and protein identifiers and covariates as columns (with additional columns 'group', 'sample' and 'subject').
    :param list covariates: list of covariates to include in the model (column in df)
    :param str subject: column with subject identifiers
    :param str group: column with group identifiers
    :param list drop_cols: column labels to be dropped from the dataframe
    :param float alpha: error rate for multiple hypothesis correction
    :param int permutations: number of permutations used to estimate false discovery rates.
    :param bool non_par: if True, normality and variance equality assumptions are checked and non-parametric test Mann Whitney U test if not passed
    :return: Pandas dataframe with columns 'identifier', 'group1', 'group2', 'mean(group1)', 'mean(group2)', 'Log2FC', 'std_error', 'tail', 't-statistics', 'posthoc pvalue', 'effsize', 'efftype', 'FC', 'rejected', 'F-statistics', 'p-value', 'correction', '-log10 p-value', and 'method'.

    Example::

        result = run_ancova(df, covariates=['age'], alpha=0.05, drop_cols=["sample",'subject'], subject='subject', group='group', permutations=50)
    """
    df = df.drop(drop_cols, axis=1)
    for cova in covariates:
        if df[cova].dtype != np.number:
            df[cova] = pd.Categorical(df[cova])
            df[cova] = df[cova].cat.codes

    pairwise_results = []
    ancova_result = []
    for col in df.columns.tolist():
        if col not in covariates and col != group:
            ancova = calculate_ancova(
                df[[group, col] + covariates], col, group=group, covariates=covariates
            )
            ancova_result.append(ancova)
            pairwise_result = pairwise_ttest_with_covariates(
                df, column=col, group=group, covariates=covariates, is_logged=is_logged
            )
            pairwise_cols = pairwise_result.columns
            pairwise_results.extend(pairwise_result.values.tolist())
    df = df.set_index([group])
    res = format_anova_table(
        df,
        ancova_result,
        pairwise_results,
        pairwise_cols,
        group,
        permutations,
        alpha,
        correction,
    )
    res["Method"] = "One-way ancova"
    res = correct_pairwise_ttest(res, alpha, correction)

    return res


def pairwise_ttest_with_covariates(df, column, group, covariates, is_logged):
    formula = "Q('%s') ~ C(Q('%s'))" % (column, group)
    for c in covariates:
        formula += " + Q('%s')" % (c)
    model = ols(formula, data=df).fit()
    pw = model.t_test_pairwise("C(Q('%s'))" % (group)).result_frame
    pw = pw.reset_index()
    groups = "|".join([re.escape(s) for s in df[group].unique().tolist()])
    regex = r"({})\-({})".format(groups, groups)
    pw["group1"] = pw["index"].apply(lambda x: re.search(regex, x).group(2))
    pw["group2"] = pw["index"].apply(lambda x: re.search(regex, x).group(1))

    means = df.groupby(group)[column].mean().to_dict()
    stds = df.groupby(group)[column].std().to_dict()
    pw["mean(group1)"] = [means[g] for g in pw["group1"].tolist()]
    pw["mean(group2)"] = [means[g] for g in pw["group2"].tolist()]
    pw["std(group1)"] = [stds[g] for g in pw["group1"].tolist()]
    pw["std(group2)"] = [stds[g] for g in pw["group2"].tolist()]
    pw = pw.drop(["pvalue-hs", "reject-hs"], axis=1)
    pw = pw.rename(columns={"t": "posthoc T-Statistics", "P>|t|": "posthoc pvalue"})

    pw = pw[
        [
            "group1",
            "group2",
            "mean(group1)",
            "std(group1)",
            "mean(group2)",
            "std(group2)",
            "posthoc T-Statistics",
            "posthoc pvalue",
            "coef",
            "std err",
            "Conf. Int. Low",
            "Conf. Int. Upp.",
        ]
    ]
    pw = complement_posthoc(pw, column, is_logged)

    return pw


def run_repeated_measurements_anova(
    df,
    alpha=0.05,
    drop_cols=["sample"],
    subject="subject",
    within="group",
    permutations=50,
    correction="fdr_bh",
    is_logged=True,
):
    """
    Performs repeated measurements anova and pairwise posthoc tests for each protein in dataframe.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str subject: column with subject identifiers
    :param str within: column with within factor identifiers
    :param list drop_cols: column labels to be dropped from the dataframe
    :param float alpha: error rate for multiple hypothesis correction
    :param int permutations: number of permutations used to estimate false discovery rates
    :return: Pandas dataframe

    Example::

        result = run_repeated_measurements_anova(df, alpha=0.05, drop_cols=['sample'], subject='subject', within='group', permutations=50)
    """
    df = df.drop(drop_cols, axis=1).dropna(axis=1)
    aov_results = []
    pairwise_results = []
    index = [within, subject]
    for col in df.columns.drop(index).tolist():
        cols = index + [col]
        aov = calculate_repeated_measures_anova(
            df[cols], column=col, subject=subject, within=within
        )
        aov_results.append(aov)
        pairwise_result = calculate_pairwise_ttest(
            df[[within, subject, col]],
            subject=subject,
            column=col,
            group=within,
            is_logged=is_logged,
        )
        pairwise_cols = pairwise_result.columns
        pairwise_results.extend(pairwise_result.values.tolist())

    df = df.set_index([subject, within])
    res = format_anova_table(
        df,
        aov_results,
        pairwise_results,
        pairwise_cols,
        within,
        permutations,
        alpha,
        correction,
    )
    res["Method"] = "Repeated measurements anova"
    res = correct_pairwise_ttest(res, alpha, correction=correction)

    return res


def run_mixed_anova(
    df,
    alpha=0.05,
    drop_cols=["sample"],
    subject="subject",
    within="group",
    between="group2",
    correction="fdr_bh",
):
    """
    In statistics, a mixed-design analysis of variance model, also known as a split-plot ANOVA, is used to test
    for differences between two or more independent groups whilst subjecting participants to repeated measures.
    Thus, in a mixed-design ANOVA model, one factor (a fixed effects factor) is a between-subjects variable and the other
    (a random effects factor) is a within-subjects variable. Thus, overall, the model is a type of mixed-effects model.
    [source:https://en.wikipedia.org/wiki/Mixed-design_analysis_of_variance]

    :param df: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str subject: column with subject identifiers
    :param str within: column with within factor identifiers
    :param str between: column with between factor identifiers
    :param list drop_cols: column labels to be dropped from the dataframe
    :param float alpha: error rate for multiple hypothesis correction
    :param int permutations: number of permutations used to estimate false discovery rates
    :return: Pandas dataframe

    Example::

        result = run_mixed_anova(df, alpha=0.05, drop_cols=['sample'], subject='subject', within='group', between='group2', permutations=50)
    """
    df = df.drop(drop_cols, axis=1).dropna(axis=1)
    aov_results = []
    index = [within, subject, between]
    for col in df.columns.drop(index).tolist():
        cols = index + [col]
        aov = calculate_mixed_anova(
            df[cols], column=col, subject=subject, within=within, between=between
        )
        aov_results.append(aov)

    res = pd.concat(aov_results)
    res = res[res["Source"] == "Interaction"]
    res = res[["identifier", "DF1", "DF2", "F", "p-unc"]]
    res.columns = ["identifier", "dfk", "dfn", "F-statistics", "pvalue"]
    _, padj = apply_pvalue_correction(
        res["pvalue"].tolist(), alpha=alpha, method=correction
    )
    res["correction"] = "FDR correction BH"
    res["padj"] = padj
    res["rejected"] = res["padj"] < alpha
    res["testing"] = "Interaction"
    res["within"] = ",".join(df[within].unique().tolist())
    res["between"] = ",".join(df[between].unique().tolist())

    return res


def format_anova_table(
    df,
    aov_results,
    pairwise_results,
    pairwise_cols,
    group,
    permutations,
    alpha,
    correction,
):
    """
    Performs p-value correction (permutation-based and FDR) and converts pandas dataframe into final format.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param list[tuple] aov_results: list of tuples with anova results (one tuple per feature).
    :param list[dataframes] pairwise_results: list of pandas dataframes with posthoc tests results
    :param str group: column with group identifiers
    :param float alpha: error rate for multiple hypothesis correction
    :param int permutations: number of permutations used to estimate false discovery rates
    :return: Pandas dataframe
    """
    columns = ["identifier", "dfk", "dfn", "F-statistics", "pvalue"]
    scores = pd.DataFrame(aov_results, columns=columns)
    scores = scores.set_index("identifier")
    corrected = False
    # FDR correction
    if permutations > 0:
        max_perm = get_max_permutations(df, group=group)
        if max_perm >= 10:
            if max_perm < permutations:
                permutations = max_perm
            observed_pvalues = scores.pvalue
            count = apply_pvalue_permutation_fdrcorrection(
                df,
                observed_pvalues,
                group=group,
                alpha=alpha,
                permutations=permutations,
            )
            scores = scores.join(count)
            scores["correction"] = "permutation FDR ({} perm)".format(permutations)
            corrected = True

    if not corrected:
        _, padj = apply_pvalue_correction(
            scores["pvalue"].tolist(), alpha=alpha, method=correction
        )
        scores["correction"] = "FDR correction BH"
        scores["padj"] = padj
        corrected = True

    res = pd.DataFrame(pairwise_results, columns=pairwise_cols).set_index("identifier")
    if not res.empty:
        res = res.join(scores[["F-statistics", "pvalue", "padj"]].astype("float"))
        res["correction"] = scores["correction"]
    else:
        res = scores
        res["log2FC"] = np.nan

    res = res.reset_index()
    res["rejected"] = res["padj"] < alpha

    if "posthoc pvalue" in res.columns:
        res["-log10 pvalue"] = [-np.log10(x) for x in res["posthoc pvalue"].values]
    else:
        res["-log10 pvalue"] = [-np.log10(x) for x in res["pvalue"].values]

    return res


def run_ttest(
    df,
    condition1,
    condition2,
    alpha=0.05,
    drop_cols=["sample"],
    subject="subject",
    group="group",
    paired=False,
    correction="fdr_bh",
    permutations=0,
    is_logged=True,
    non_par=False,
):
    """
    Runs t-test (paired/unpaired) for each protein in dataset and performs permutation-based (if permutations>0) or Benjamini/Hochberg (if permutations=0) multiple hypothesis correction.

    :param df: pandas dataframe with samples as rows and protein identifiers as columns (with additional columns 'group', 'sample' and 'subject').
    :param str condition1: first of two conditions of the independent variable
    :param str condition2: second of two conditions of the independent variable
    :param str subject: column with subject identifiers
    :param str group: column with group identifiers (independent variable)
    :param list drop_cols: column labels to be dropped from the dataframe
    :param bool paired: paired or unpaired samples
    :param str correction: method of pvalue correction see apply_pvalue_correction for methods
    :param float alpha: error rate for multiple hypothesis correction
    :param int permutations: number of permutations used to estimate false discovery rates.
    :param bool is_logged: data is log-transformed
    :param bool non_par: if True, normality and variance equality assumptions are checked and non-parametric test Mann Whitney U test if not passed
    :return: Pandas dataframe with columns 'identifier', 'group1', 'group2', 'mean(group1)', 'mean(group2)', 'std(group1)', 'std(group2)', 'Log2FC', 'FC', 'rejected', 'T-statistics', 'p-value', 'correction', '-log10 p-value', and 'method'.

    Example::

        result = run_ttest(df, condition1='group1', condition2='group2', alpha = 0.05, drop_cols=['sample'], subject='subject', group='group', paired=False, correction='fdr_bh', permutations=50)
    """
    columns = [
        "T-statistics",
        "pvalue",
        "mean_group1",
        "mean_group2",
        "std(group1)",
        "std(group2)",
        "log2FC",
        "test",
    ]
    df = df.set_index(group)
    df = df.drop(drop_cols, axis=1)
    method = "Unpaired t-test"
    if non_par:
        method = "Unpaired t-Test and Mann-Whitney U test"

    if paired:
        df = df.reset_index().set_index([group, subject])
        method = "Paired t-test"
    else:
        if subject is not None:
            df = df.drop([subject], axis=1)

    scores = df.T.apply(
        func=calculate_ttest,
        axis=1,
        result_type="expand",
        args=(condition1, condition2, paired, is_logged, non_par),
    )
    scores.columns = columns
    scores = scores.dropna(how="all")

    corrected = False
    # FDR correction
    if permutations > 0:
        max_perm = get_max_permutations(df, group=group)
        if max_perm >= 10:
            if max_perm < permutations:
                permutations = max_perm
            observed_pvalues = scores.pvalue
            count = apply_pvalue_permutation_fdrcorrection(
                df,
                observed_pvalues,
                group=group,
                alpha=alpha,
                permutations=permutations,
            )
            scores = scores.join(count)
            scores["correction"] = "permutation FDR ({} perm)".format(permutations)
            corrected = True

    if not corrected:
        rejected, padj = apply_pvalue_correction(
            scores["pvalue"].tolist(), alpha=alpha, method=correction
        )
        scores["correction"] = "FDR correction BH"
        scores["padj"] = padj
        scores["rejected"] = rejected
        corrected = True

    scores["group1"] = condition1
    scores["group2"] = condition2
    if is_logged:
        scores["FC"] = scores["log2FC"].apply(lambda x: np.power(2, x))
    else:
        scores = scores.rename(columns={"log2FC": "FC"})

    scores["-log10 pvalue"] = [
        -np.log10(x) if x != 0 else -np.log10(alpha) for x in scores["pvalue"].values
    ]
    scores["Method"] = method
    scores.index.name = "identifier"
    scores = scores.reset_index()

    return scores


def calculate_pvalue_from_tstats(tstat, dfn):
    """
    Calculate two-tailed p-values from T- or F-statistics.

    tstat: T/F distribution
    dfn: degrees of freedrom *n* (values) per protein (keys), i.e. number of obervations - number of groups (dict)
    """
    pval = scipy.stats.t.sf(np.abs(tstat), dfn) * 2

    return pval


def run_two_way_anova(
    df, drop_cols=["sample"], subject="subject", group=["group", "secondary_group"]
):
    """
    Run a 2-way ANOVA when data['secondary_group'] is not empty

    :param df: processed pandas dataframe with samples as rows, and proteins and groups as columns.
    :param list drop_cols: column names to drop from dataframe
    :param str subject: column name containing subject identifiers.
    :param list group: column names corresponding to independent variable groups
    :return: Two dataframes, anova results and residuals.

    Example::

        result = run_two_way_anova(data, drop_cols=['sample'], subject='subject', group=['group', 'secondary_group'])
    """
    data = df.copy()
    factorA, factorB = group
    data = data.set_index([subject] + group)
    data = data.drop(drop_cols, axis=1)
    data.columns = data.columns.str.replace(r"-", "_")

    aov_result = []
    residuals = {}
    for col in data.columns:
        model = ols(
            "{} ~ C({})*C({})".format(col, factorA, factorB),
            data[col].reset_index().sort_values(group, ascending=[True, False]),
        ).fit()
        aov_table = sm.stats.anova_lm(model, typ=2)
        eta_squared(aov_table)
        omega_squared(aov_table)
        for i in aov_table.index:
            if i != "Residual":
                t, p, eta, omega = aov_table.loc[
                    i, ["F", "PR(>F)", "eta_sq", "omega_sq"]
                ]
                protein = col.replace("_", "-")
                aov_result.append((protein, i, t, p, eta, omega))
        residuals[col] = model.resid

    anova_df = pd.DataFrame(
        aov_result,
        columns=[
            "identifier",
            "source",
            "F-statistics",
            "pvalue",
            "eta_sq",
            "omega_sq",
        ],
    )
    anova_df = anova_df.set_index("identifier")
    anova_df = anova_df.dropna(how="all")

    return anova_df, residuals


def eta_squared(aov):
    """
    Calculates the effect size using Eta-squared.

    :param aov: pandas dataframe with anova results from statsmodels.
    :return: Pandas dataframe with additional Eta-squared column.
    """
    aov["eta_sq"] = "NaN"
    aov["eta_sq"] = aov[:-1]["sum_sq"] / sum(aov["sum_sq"])
    return aov


def omega_squared(aov):
    """
    Calculates the effect size using Omega-squared.

    :param aov: pandas dataframe with anova results from statsmodels.
    :return: Pandas dataframe with additional Omega-squared column.
    """
    mse = aov["sum_sq"][-1] / aov["df"][-1]
    aov["omega_sq"] = "NaN"
    aov["omega_sq"] = (aov[:-1]["sum_sq"] - (aov[:-1]["df"] * mse)) / (
        sum(aov["sum_sq"]) + mse
    )
    return aov



import rarfile


# ! unrar must be installed in the system
def unrar(filepath, to):
    """
    Decompress RAR file
    :param str filepath: path to rar file
    :param str to: where to extract all files

    """
    try:
        with rarfile.RarFile(filepath) as opened_rar:
            print("Extracting files: {}".format(opened_rar.namelist()))
            opened_rar.extractall(to)
    except Exception as err:
        print("Error: {}. Could not unrar file {}".format(filepath, err))
        raise



# %% [metadata]
# # PXD010372
# %%
from pathlib import Path

import pandas as pd

import acore.io

# %%
fpath_meta = "meta.csv"  # in same folder as this script
furl_pg = "https://ars.els-cdn.com/content/image/1-s2.0-S0092867418311668-mmc2.xlsx"

# %% [markdown]
# Where to save the dat

# %%
dir_proc = Path("processed")
dir_proc.mkdir(exist_ok=True)

# %%
local_filename = Path(furl_pg).name
acore.io.download_file(furl_pg, local_filename)

# %%
meta_patients = pd.read_csv(fpath_meta, index_col=0)
meta_patients

# %%
data = pd.read_excel(
    local_filename,
    sheet_name="SupplementaryTable2_PatientProt",
)
data

# %% [markdown]
# We will use the first protein in the a protein group as identifier,
# which we verify to be unique.

# %%
data["first_prot"] = data["Majority protein Ids"].str.split(";").str[0]
data["first_prot"].nunique() == data["Majority protein Ids"].nunique()
data = data.set_index("first_prot")
assert data.index.is_unique
data

# %% [markdown]
# Filter intensity values for patients

# %%
pgs = data.filter(like="Patient")
pgs

# %% [markdown]
# There are two measurements for patient 11 in the data: 11 and 11B. In the methods of
# the paper it is stated:
#
# "We required a minimum peptide ratio count of 1 to report a quantitative readout
# and averaged the results from duplicate measurements of the same sample."
#
# We will do this manually for patient 11 measurements.

# %%
pgs.filter(like="Patient11").describe()

# %%
pgs = pgs.assign(Patient11=lambda df: df.filter(like="Patient11").mean(axis=1)).drop(
    ["Patient11B"], axis=1
)

# %%
patient_int_map = (
    pgs.columns.to_frame()
    .squeeze()
    .str.extract(r"(\d+)")
    .squeeze()
    .astype(int)
    .to_frame("Patient")
    .reset_index()
    .set_index("Patient")
    .rename({"index": "Patient_id"}, axis=1)
)
patient_int_map

# %%
meta_patients = meta_patients.join(patient_int_map).set_index(
    "Patient_id",
)
fname = dir_proc / "meta_patients.csv"
meta_patients.to_csv(fname)
print("Saved to ", fname)
meta_patients

# %%
meta_pgs = data.drop(pgs.columns, axis=1).drop("Patient11B", axis=1)
fname = dir_proc / "meta_pgs.csv"
meta_pgs.to_csv(fname)
print("Saved to ", fname)
meta_pgs

# %%
meta_pgs.describe(exclude="number")

# %%
mask = meta_pgs["Potential contaminant"].isna()
mask.value_counts()

# %%
pgs = pgs.loc[mask].T
fname = dir_proc / "omics.csv"
pgs.to_csv(fname)
print("Saved to ", fname)
pgs

# %% [markdown]
# Need to decide how to handle the duplicated patient measurement for patient 11.



import pandas as pd
import sklearn.decomposition


# ! also a version in exploratory analysis
def run_pca(
    df_wide: pd.DataFrame, n_components: int = 2
) -> tuple[pd.DataFrame, sklearn.decomposition.PCA]:
    """Run PCA on DataFrame and return result.

    Parameters
    ----------
    df : pd.DataFrame
        DataFrame in wide format to fit features on.
    n_components : int, optional
        Number of Principal Components to fit, by default 2

    Returns
    -------
    Tuple[pd.DataFrame, PCA]
        principal compoments of DataFrame with same indices as in original DataFrame,
        and fitted PCA model of sklearn
    """
    n_comp_max = None
    if n_components is not None:
        n_comp_max = min(df_wide.shape)
        n_comp_max = min(n_comp_max, n_components)
    pca = sklearn.decomposition.PCA(n_components=n_comp_max)
    pcs = pca.fit_transform(df_wide)
    cols = [
        f"principal component {i+1} ({var_explained*100:.2f} %)"
        for i, var_explained in enumerate(pca.explained_variance_ratio_)
    ]
    pcs = pd.DataFrame(pcs, index=df_wide.index, columns=cols)
    return pcs, pca



import community
import networkx as nx
import pandas as pd
import snf
from sklearn import cluster
from sklearn.cluster import AffinityPropagation

from . import utils


def get_network_communities(graph, args):
    """
    Finds communities in a graph using different methods. For more information on the methods visit:

        - https://networkx.github.io/documentation/latest/reference/algorithms/generated/networkx.algorithms.community.modularity_max.greedy_modularity_communities.html
        - https://networkx.github.io/documentation/networkx-2.0/reference/algorithms/generated/networkx.algorithms.community.asyn_lpa.asyn_lpa_communities.html
        - https://networkx.github.io/documentation/latest/reference/algorithms/generated/networkx.algorithms.community.centrality.girvan_newman.html
        - https://networkx.github.io/documentation/latest/reference/generated/networkx.convert_matrix.to_pandas_adjacency.html

    :param graph graph: networkx graph
    :param dict args: config file arguments
    :return: Dictionary of nodes and which community they belong to (from 0 to number of communities).
    """
    if "communities_algorithm" not in args:
        args["communities_algorithm"] = "louvain"

    if args["communities_algorithm"] == "louvain":
        communities = get_louvain_partitions(graph, args["values"])
    elif args["communities_algorithm"] == "greedy_modularity":
        gcommunities = nx.algorithms.community.greedy_modularity_communities(
            graph, weight=args["values"]
        )
        communities = utils.generator_to_dict(gcommunities)
    elif args["communities_algorithm"] == "asyn_label_propagation":
        gcommunities = nx.algorithms.community.label_propagation.asyn_lpa_communities(
            graph, args["values"]
        )
        communities = utils.generator_to_dict(gcommunities)
    elif args["communities_algorithm"] == "girvan_newman":
        gcommunities = nx.algorithms.community.girvan_newman(
            graph, most_valuable_edge=most_central_edge
        )
        communities = utils.generator_to_dict(gcommunities)
    elif args["communities_algorithm"] == "affinity_propagation":
        adjacency = nx.to_pandas_adjacency(graph, weight="width")
        nodes = list(adjacency.columns)
        communities = AffinityPropagation().fit(adjacency.values).labels_
        communities = {nodes[i]: communities[i] for i in range(len(communities))}

    return communities


def get_snf_clusters(data_tuples, num_clusters=None, metric="euclidean", k=5, mu=0.5):
    """
    Cluster samples based on Similarity Network Fusion (SNF) (ref: https://www.ncbi.nlm.nih.gov/pubmed/24464287)

    :param df_tuples: list of (dataset,metric) tuples
    :param index: how the datasets can be merged (common columns)
    :param num_clusters: number of clusters to be identified, if None, the algorithm finds the best number based on SNF algorithm (recommended)
    :param distance_metric: distance metric used to calculate the sample similarity network
    :param k: number of neighbors used to measure local affinity (KNN)
    :param mu: normalization factor to scale similarity kernel when constructing affinity matrix
    :return tuple: 1) fused_aff: affinity clustered samples, 2) fused_labels: cluster labels,
                    3) num_clusters: number of clusters, 4) silhouette: average silhouette score
    """
    affinities = []
    for d, m in data_tuples:
        affinities += [snf.make_affinity(d, metric=m, K=k, mu=mu)]
    fused_aff = snf.snf(affinities, K=k)
    if num_clusters is None:
        num_clusters, second = snf.get_n_clusters(fused_aff)
    fused_labels = cluster.spectral_clustering(fused_aff, n_clusters=num_clusters)
    fused_labels = [i + 1 for i in fused_labels]
    silhouette = snf.metrics.silhouette_score(fused_aff, fused_labels)

    return (fused_aff, fused_labels, num_clusters, silhouette)


def most_central_edge(G):
    """
    Compute the eigenvector centrality for the graph G, and finds the highest value.

    :param graph G: networkx graph
    :return: Highest eigenvector centrality value.
    :rtype: float
    """
    centrality = nx.eigenvector_centrality_numpy(G, weight="width")

    return max(centrality, key=centrality.get)


def get_louvain_partitions(G, weight):
    """
    Computes the partition of the graph nodes which maximises the modularity (or try..) using the Louvain heuristices. For more information visit https://python-louvain.readthedocs.io/en/latest/api.html.

    :param graph G: networkx graph which is decomposed.
    :param str weight: the key in graph to use as weight.
    :return: The partition, with communities numbered from 0 to number of communities.
    :rtype: dict
    """
    partition = community.best_partition(G, weight=weight)

    return partition


def run_snf(
    df_dict,
    index,
    num_clusters=None,
    distance_metric="euclidean",
    k_affinity=5,
    mu_affinity=0.5,
):
    """
    Runs Similarity Network Fusion: integration of multiple omics datasets to identify
    similar samples (clusters) (ref: https://www.ncbi.nlm.nih.gov/pubmed/24464287).
    We make use of the pyton version SNFpy (https://github.com/rmarkello/snfpy)

    :param df_dict: dictionary of datasets to be used (i.e {'rnaseq': rnaseq_data, 'proteomics': proteomics_data})
    :param index: how the datasets can be merged (common columns)
    :param num_clusters: number of clusters to be identified, if None, the algorithm finds the best number based on SNF algorithm (recommended)
    :param distance_metric: distance metric used to calculate the sample similarity network
    :param k_affinity: number of neighbors used to measure local affinity (KNN)
    :param mu_ffinity: normalization factor to scale similarity kernel when constructing affinity matrix
    :return tuple: 1) feature_df: SNF features and mutual information score (MIscore), 2) fused_aff: adjacent similarity matrix, 3)fused_labels: cluster labels,
                    4) silhouette: silhouette score
    """
    datasets = []
    dataset_labels = []
    common_samples = set()
    for dtype in df_dict:
        dataset_labels.append(dtype)
        df = df_dict[dtype]
        df = df.set_index(index)
        datasets.append(df)
        if len(common_samples) > 1:
            common_samples = common_samples.intersection(df.index)
        else:
            common_samples.update(df.index.tolist())

    data_tuples = [
        (d.loc[list(common_samples)].values, distance_metric) for d in datasets
    ]

    fused_aff, fused_labels, num_clusters, silhouette_score = get_snf_clusters(
        data_tuples, num_clusters, metric=distance_metric, k=k_affinity, mu=mu_affinity
    )

    fused_labels = pd.DataFrame(fused_labels, index=common_samples, columns=["cluster"])

    snf_features = snf.metrics.rank_feature_by_nmi(
        data_tuples, fused_aff, K=k_affinity, mu=mu_affinity, n_clusters=num_clusters
    )

    feature_df = pd.DataFrame(columns=["MIscore"])
    indexes = [df.columns for df in datasets]
    i = 0
    for dtype in snf_features:
        df = pd.DataFrame(dtype, index=indexes[i], columns=["MIscore"]).sort_values(
            by="MIscore", ascending=False
        )
        df["dataset"] = dataset_labels[i]
        i += 1
        feature_df = feature_df.append(df)

    feature_df = feature_df.sort_values(by="MIscore", ascending=False)

    return feature_df, fused_aff, fused_labels, silhouette_score



# %%
from pathlib import Path

import acore.io as aio

id = "PXD008541"
fname = "SearchEngineResults_secretome.zip.rar"
folder_downloads = Path("downloaded")
folder_unzipped = Path("unzipped")

# %%
ret = aio.download_PRIDE_data(id, fname, to="downloaded")
ret

# %% [markdown]
# see ftp link for PRIDE project

# %%
ret["_links"]["datasetFtpUrl"]["href"]


# %%
# ! you need a system installation of a rar archive tool
aio.unrar(filepath=folder_downloads / fname, to=folder_unzipped)

# %%
list(folder_unzipped.iterdir())

# %%



import pandas as pd
from lifelines import KaplanMeierFitter, NelsonAalenFitter
from lifelines.statistics import multivariate_logrank_test


def get_data_ready_for_km(dfs_dict, args):
    kmdf = None
    for df_key in dfs_dict:
        if df_key == "clinical":
            kmdf = dfs_dict[df_key]
        else:
            df = dfs_dict[df_key]
            if "marker" in args and "index_col" in args:
                how = "half"
                value = None
                if "how" in args:
                    how = args["how"]
                if "value" in args:
                    value = args["value"]
                mdf = group_data_based_on_marker(
                    df, args["marker"], args["index_col"], how, value
                )

    if kmdf is not None:
        if "index_col" in args and args["index_col"] in kmdf:
            index_col = args["index_col"]
            kmdf = kmdf.set_index(index_col).join(mdf.set_index(index_col), how="inner")

    return kmdf


def group_data_based_on_marker(df, marker, index_col, how, value):
    mdf = pd.DataFrame()
    if index_col is not None and marker is not None:
        if index_col in df and marker in df:
            mdf = df[[marker, index_col]]
            if how == "cutoff":
                mdf["new_grouping"] = mdf.apply(
                    lambda row: (
                        str(marker) + "+" if row[marker] >= value else str(marker) + "-"
                    )
                )
            elif how == "top" or how == "top%":
                mdf = mdf.sort_values(by=marker, ascending=False)
                num_values = len(mdf[marker].values.tolist())
                if how == "top%":
                    value = int(num_values * value / 100)
                if value < num_values:
                    labels = [str(marker) + "+"] * value
                    labels.extend([str(marker) + "-"] * (num_values - value))
                else:
                    print(
                        "Invalid value provided. Exceeded maximun number of samples {}"
                        .format(num_values)
                    )
                mdf["new_grouping"] = labels
            else:
                print(
                    "Grouping method {} not implemented. Try with 'cutoff' or 'top'"
                    .format(how)
                )

    return mdf


def run_km(data, time_col, event_col, group_col, args={}):
    kmdf = None
    kmf = pd.DataFrame()
    summary = None
    if isinstance(data, dict):
        kmdf = get_data_ready_for_km(data, args)
        group_col = "new_grouping"
    elif isinstance(data, pd.DataFrame):
        kmdf = data

    if kmdf is not None:
        kmf, summary = get_km_results(kmdf, group_col, time_col, event_col)

    return kmf, summary


def get_km_results(df, group_col, time_col, event_col):
    models = []
    summary_ = None
    summary_result = None
    df = df[[event_col, time_col, group_col]].dropna()
    df[event_col] = df[event_col].astype("category")
    df[event_col] = df[event_col].cat.codes
    df[time_col] = df[time_col].astype("float")
    if not df.empty:
        for name, grouped_df in df.groupby(group_col):
            kmf = KaplanMeierFitter()
            t = grouped_df[time_col]
            e = grouped_df[event_col]
            kmf.fit(
                t, event_observed=e, label=name + " (N=" + str(len(t.tolist())) + ")"
            )
            models.append(kmf)

        summary_ = multivariate_logrank_test(
            df[time_col].tolist(),
            df[group_col].tolist(),
            df[event_col].tolist(),
            alpha=99,
        )

    if summary_ is not None:
        summary_result = "Multivariate logrank test: pval={}, t_statistic={}".format(
            summary_.p_value, summary_._test_statistic
        )

    return models, summary_result


def get_hazard_ratio_results(df, group_col, time_col, event_col):
    models = []
    df = df[[event_col, time_col, group_col]].dropna()
    df[event_col] = df[event_col].astype("category")
    df[event_col] = df[event_col].cat.codes
    df[time_col] = df[time_col].astype("float")
    if not df.empty:
        for name, grouped_df in df.groupby(group_col):
            hr = NelsonAalenFitter()
            t = grouped_df[time_col]
            e = grouped_df[event_col]
            hr.fit(
                t, event_observed=e, label=name + " (N=" + str(len(t.tolist())) + ")"
            )
            models.append(hr)

    return models



"""Downlaod data from PRIDE database."""

import requests

from .ftp import download_from_ftp


def download_PRIDE_data(
    pxd_id, file_name, to=".", user="", password="", date_field="publicationDate"
) -> dict:
    """
    This function downloads a project file from the PRIDE repository. To see more of the
    pride API, have a look at
    https://www.ebi.ac.uk/pride/ws/archive/v3/webjars/swagger-ui/index.html
    or EBI's commandline tool pridepy
    https://github.com/PRIDE-Archive/pridepy

    :param str pxd_id: PRIDE project identifier (id. PXD013599).
    :param str file_name: name of the file to dowload
    :param str to: local directory where the file should be downloaded
    :param str user: username to access biomedical database server if required.
    :param str password: password to access biomedical database server if required.
    :param str date_field: projects deposited in PRIDE are search based on date, either
        submissionData or publicationDate (default)
    """
    ftp_pride = (  # same as data ["_links"]["datasetFtpUrl"]["href"]
        "ftp://ftp.pride.ebi.ac.uk/pride/data/archive/YEAR/MONTH/PXDID/FILE_NAME"
    )
    url_pride_api = "http://www.ebi.ac.uk/pride/ws/archive/v3/projects/" + pxd_id
    data = None
    r = requests.get(url_pride_api, timeout=60)
    data = r.json()
    submission_date = data[date_field]
    year, month, _ = submission_date.split("-")

    ftp_url = (
        ftp_pride.replace("YEAR", year)
        .replace("MONTH", month)
        .replace("PXDID", pxd_id)
        .replace("FILE_NAME", file_name)
    )
    download_from_ftp(ftp_url, user, password, to, file_name)

    return data



"""Module normalization for analysis. This module provides functions to normalize data for analysis.

The higher-level convience functions are `normalize_data` and `normalize_data_by_group`.
Combat normalization was added using the inmoose package.

The actual normalization functions are in strategies.py.
"""

from __future__ import annotations

import pandas as pd

# PyPI pycombat as described in the docstring was never used:
# https://pypi.org/project/pycombat/
# https://github.com/epigenelabs/inmoose is the update from combat.pycombat on PyPI
# from combat.pycombat import pycombat
from inmoose.pycombat import pycombat_norm

from .strategies import (
    linear_normalization,
    median_normalization,
    median_polish_normalization,
    median_zero_normalization,
    quantile_normalization,
    zscore_normalization,
)

__all__ = ["combat_batch_correction", "normalize_data", "normalize_data_per_group"]


def combat_batch_correction(
    data: pd.DataFrame,
    batch_col: str,
    # index_cols: list[str],
) -> pd.DataFrame:
    """
    This function corrects processed data for batch effects. For more information visit:
    https://github.com/epigenelabs/inmoose

    :param data: pandas.DataFrame with samples as rows and protein identifiers as columns.
    :param batch_col: column with the batch identifiers
    :return: pandas.DataFrame with samples as rows and protein identifiers as columns.
    Example::
        result = combat_batch_correction(
                    data,
                    batch_col="batch",
                    index_cols=["subject", "sample", "group"],
                )

    """
    # :param index_cols: list of columns that don't need to be corrected (i.e group)
    df_corrected = pd.DataFrame()
    # index_cols = [c for c in index_cols if c != batch_col]
    # data = data.set_index(index_cols)  # ? should this not be provided directly as data
    df = data.drop(batch_col, axis=1)
    df_numeric = df.select_dtypes("number")
    num_batches = len(data[batch_col].unique())
    if df_numeric.empty:
        raise ValueError("No numeric columns found in data.")
    if not num_batches > 1:
        raise ValueError("Only one batch found in data.")
    info_cols = df.columns.difference(df_numeric.columns)
    df_corrected = pd.DataFrame(
        pycombat_norm(df_numeric.T, data[batch_col]).T,
        index=df.index,
    )
    df_corrected = df_corrected.join(df[info_cols])
    # df_corrected = df_corrected  # .reset_index()  # ? would also not reset index here

    return df_corrected


def normalize_data_per_group(
    data: pd.DataFrame,
    group: str | int | list[str | int],
    method: str = "median",
    normalize: str = None,
) -> pd.DataFrame:
    """
    This function normalizes the data by group using the selected method

    :param data: DataFrame with the data to be normalized (samples x features)
    :param group_col: Column containing the groups, passed to pandas.DataFrame.groupby
    :param str method: normalization method to choose among: median_polish, median,
                        quantile, linear
    :param str normalize: whether the normalization should be done by 'features' (columns) or 'samples' (rows) (default None)
    :return: pandas.DataFrame.

    Example::

        result = normalize_data_per_group(data, group='group' method='median')
    """
    ndf = pd.DataFrame(columns=data.columns)
    for _, gdf in data.groupby(group):
        norm_group = normalize_data(gdf, method=method, normalize=normalize)
        ndf = ndf.append(norm_group)

    return ndf


def normalize_data(
    data: pd.DataFrame,
    method: str = "median",
    normalize: str = None,
):
    """
    This function normalizes the data using the selected method. Normalizes only nummeric
    data, but keeps the non-numeric columns in the output DataFrame.

    :param data: DataFrame with the data to be normalized (samples x features)
    :param str method: normalization method to choose among: median (default),
                       median_polish, median_zero, quantile, linear, zscore
    :param str normalize: whether the normalization should be done by 'features' (columns)
                          or 'samples' (rows) (default None)
    :return: pandas.DataFrame.

    Example::

        result = normalize_data(data, method='median_polish')
    """
    numeric_cols = data.select_dtypes(
        include=["int64", "float64"]
    )  # ! too restrictive?
    non_numeric_cols = data.select_dtypes(
        exclude=["int64", "float64"]
    )  # ! too restrictive?
    if numeric_cols.empty:
        raise ValueError("No numeric columns found in data.")

    if method == "median_polish":
        norm_data = median_polish_normalization(numeric_cols, max_iter=250)
    elif method == "median_zero":
        norm_data = median_zero_normalization(numeric_cols, normalize)
    elif method == "median":
        norm_data = median_normalization(numeric_cols, normalize)
    elif method == "quantile":
        norm_data = quantile_normalization(numeric_cols)
    elif method == "linear":
        norm_data = linear_normalization(numeric_cols, method="l1", normalize=normalize)
    elif method == "zscore":
        norm_data = zscore_normalization(numeric_cols, normalize)
    else:
        raise ValueError(
            "Invalid normalization method. Should be one of:"
            " 'median', 'median_polish', 'median_zero', 'quantile', 'linear', 'zscore'"
        )

    if non_numeric_cols is not None and not non_numeric_cols.empty:
        norm_data = norm_data.join(non_numeric_cols)

    return norm_data



import platform
import unittest

import pandas as pd

import acore.exploratory_analysis as ea


class TestExtractMissing(unittest.TestCase):
    def setUp(self):
        self.data = pd.DataFrame(
            {
                "group": ["A", "A", "B", "B"],
                "sample": ["S1", "S2", "S3", "S4"],
                "protein1": [1.4, 2.2, None, 4.2],
                "protein2": [5.6, None, None, 8.1],
                "protein3": [9.1, 10.01, 11.2, 12.9],
            }
        )

    def test_extract_number_missing(self):
        min_valid = 2
        expected_result = ["protein1", "protein3"]

        result = ea.extract_number_missing(
            self.data, min_valid, drop_cols=["sample"], group="group"
        )

        self.assertEqual(result, expected_result)

    def test_extract_percentage_missing(self):
        missing_max = 0.2
        expected_result = ["protein1", "protein3"]

        result = ea.extract_percentage_missing(
            self.data, missing_max, drop_cols=["sample"], group="group"
        )

        self.assertEqual(result, expected_result)


class TestDimensionalityReduction(unittest.TestCase):
    def setUp(self):
        self.data = pd.DataFrame(
            {
                "group": ["A", "A", "B", "B"],
                "protein1": [1.4, 2.2, 5.3, 4.2],
                "protein2": [5.6, 0.3, 2.1, 8.1],
                "protein3": [9.1, 10.01, 11.2, 12.9],
            }
        )

    def test_run_pca(self):
        components = 2
        expected_result = (
            pd.DataFrame(
                {
                    "group": ["A", "A", "B", "B"],
                    "x": [0.877154, -3.883458, -1.550791, 4.557095],
                    "y": [-2.815737, -0.420979, 2.278209, 0.958507],
                }
            ),
            pd.DataFrame(
                {
                    "x": [0.958489, 0.092382, 0.269749],
                    "y": [-0.235412, 0.790179, 0.565861],
                    "value": [0.986975, 0.795561, 0.626868],
                },
                index=["protein2", "protein1", "protein3"],
            ),
            pd.Series([0.71760534, 0.26139782]),
        )
        exp_dict = {"x_title": "PC1 (0.72)", "y_title": "PC2 (0.26)", "group": "group"}
        result, annotation = ea.run_pca(
            self.data,
            drop_cols=[],
            annotation_cols=[],
            group="group",
            components=components,
            dropna=True,
        )
        assert annotation == exp_dict
        # results: tuple of 3 DFs: components, loadings, explained_variance
        pd.testing.assert_frame_equal(result[0], expected_result[0], check_exact=False)
        pd.testing.assert_frame_equal(result[1], expected_result[1], check_exact=False)
        pd.testing.assert_series_equal(
            pd.Series(result[2]),
            expected_result[2],
            check_exact=False,
            check_dtype=False,
        )

    def test_run_tsne(self):
        components = 2
        # ! MAC ARM64 yields these results (using the similar installation)
        expected_result = pd.DataFrame(
            {
                "group": ["A", "A", "B", "B"],
                "x": [-113.341728, 36.020717, 57.2992514, -134.729141],
                "y": [131.169082, -59.616630, 110.601203, -39.086254],
            }
        )
        # ! Ubuntu x64 yields these results:
        if not platform.machine() == "arm64":
            expected_result = pd.DataFrame(
                {
                    "group": ["A", "A", "B", "B"],
                    "x": [
                        -27.375560760498047,
                        -7.526909828186035,
                        -28.731882095336914,
                        -8.872751235961914,
                    ],
                    "y": [
                        1.547648310661316,
                        0.17019487917423248,
                        -18.321491241455078,
                        -19.713821411132812,
                    ],
                }
            )
        result, _ = ea.run_tsne(
            self.data,
            drop_cols=["sample", "subject"],
            group="group",
            components=components,
            perplexity=3,
            n_iter=1000,
            init="pca",
            dropna=True,
        )
        result = result["tsne"]
        pd.testing.assert_frame_equal(
            result, expected_result, check_exact=False, check_dtype=False
        )

    def test_run_umap(self):
        _ = ea.run_umap(
            self.data,
            drop_cols=["sample", "subject"],
            group="group",
            n_neighbors=10,
            min_dist=0.3,
            metric="cosine",
            dropna=True,
        )

        self.assertEqual(1, 1)


if __name__ == "__main__":
    unittest.main()



from urllib import error

import pandas as pd
from Bio import Entrez, Medline

# TODO: This should probably be changed to the email of the person installing ckg?
Entrez.email = "kg@dtu.dk"


def getMedlineAbstracts(idList):
    fields = {
        "TI": "title",
        "AU": "authors",
        "JT": "journal",
        "DP": "date",
        "MH": "keywords",
        "AB": "abstract",
        "PMID": "PMID",
    }
    pubmedUrl = "https://www.ncbi.nlm.nih.gov/pubmed/"
    abstracts = pd.DataFrame()
    try:
        handle = Entrez.efetch(
            db="pubmed", id=idList, rettype="medline", retmode="json"
        )
        records = Medline.parse(handle)
        results = []
        for record in records:
            aux = {}
            for field in fields:
                if field in record:
                    aux[fields[field]] = record[field]

            if "PMID" in aux:
                aux["url"] = pubmedUrl + aux["PMID"]
            else:
                aux["url"] = ""
            results.append(aux)

        abstracts = pd.DataFrame.from_dict(results)
    except error.URLError as e:
        print("URLError: Request to Bio.Entrez failed. Error: {}".format(e))
    except error.HTTPError as e:
        print("HTTPError: Request to Bio.Entrez failed. Error: {}".format(e))
    except Exception as e:
        print("Request to Bio.Entrez failed. Error: {}".format(e))

    return abstracts


def get_publications_abstracts(
    data,
    publication_col="publication",
    join_by=["publication", "Proteins", "Diseases"],
    index="PMID",
):
    """
    Accesses NCBI PubMed over the WWW and retrieves the abstracts corresponding
    to a list of one or more PubMed IDs.

    :param data: pandas dataframe of diseases and publications linked to a list of
                proteins (columns: 'Diseases', 'Proteins', 'linkout' and 'publication').
    :param str publication_col: column label containing PubMed ids.
    :param list join_by: column labels to be kept from the input dataframe.
    :param str index: column label containing PubMed ids from the NCBI retrieved data.
    :return: Pandas dataframe with publication information and columns 'PMID', 'abstract',
     'authors', 'date', 'journal', 'keywords', 'title', 'url', 'Proteins' and 'Diseases'.

    Example::

        result = get_publications_abstracts(data,
                    publication_col='publication',
                    join_by=['publication','Proteins','Diseases'],
                    index='PMID')
    """
    abstracts = pd.DataFrame()
    if not data.empty:
        abstracts = getMedlineAbstracts(
            list(data.reset_index()[publication_col].unique())
        )
        if not abstracts.empty:
            abstracts = abstracts.set_index(index)
            abstracts = abstracts.join(
                data.reset_index()[join_by].set_index(publication_col)
            )
            abstracts.index.name = index
            abstracts = abstracts.reset_index()
    return abstracts



# ---
# jupyter:
#   jupytext:
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Use proteomics data from PRIDE Data (adipose tissue)
# This notebook shows how `acore` can be used to download data from
# the Proteomics Identifications Database - PRIDE -
# ([ebi.ac.uk/pride/](https://www.ebi.ac.uk/pride/))
# and parse the data to be used in the analytics core
# and quickly formated to start analyzing them with the functionality in the analytics core.
#
# > based on CKG recipe: [Download PRIDE Data](https://ckg.readthedocs.io/en/latest/notebooks/recipes/Download_PRIDE_data.html)

# %% tags=["hide-output"]
# %pip install acore

# %% tags=["hide-input"]
from pathlib import Path

import numpy as np
import pandas as pd

import acore.io

# %% [markdown]
# ## Parameters
# Specify the PRIDE identifier and file to be downloaded
# and where to store intermediate files.

# %% tags=["parameters"]
pxd_id: str = "PXD008541"  # PRIDE identifier
fname = "SearchEngineResults_secretome.zip.rar"  # file to download
folder_downloads = Path("downloaded")  # folder to download the file
folder_unzipped = Path("unzipped")  # folder to uncompress the file

# %% [markdown]
# ## Specify the PRIDE identifier and file to be downloaded
#
# We can use functionality in `acore` to directly download data files from EBI's
# PRIDE database [ebi.ac.uk/pride/](https://www.ebi.ac.uk/pride/).
# For that you just need to specify the
# PRIDE identifier for the project (`PXD_...`) and the name of the file to download.
# In this case, the project identifier is `PXD008541` and the file we will use
# is `SearchEngineResults_secretome.zip.rar`,
# a RAR compressed file with the output files from MaxQuant.

# %%
ret = acore.io.download_PRIDE_data(pxd_id=pxd_id, file_name=fname, to=folder_downloads)
ret["acore_downloaded_file"] = folder_downloads / fname
ret

# %% [markdown]
# ## Decompress rar File
# Pride results are compressed by the researcher themself, so many different file
# formats can be found. Here it was stored as a RAR archive. You will need to have
# a system installation of a rar archive tool to decompress the file, find it
# via [google](https://www.google.com/search?q=unrar+tool&oq=unrar+tool).

# %%
# # ! you need a system installation of a rar archive tool
acore.io.unrar(filepath=ret["acore_downloaded_file"], to=folder_unzipped)

# %% [markdown]
# The list of files within the compressed folder

# %%
list(folder_unzipped.iterdir())

# %% [markdown]
# ## Read and clean the data
# We use the proteinGroups file that contains the proteomics data processed
# using MaxQuant software.

# %%
fpath_proteinGroups = folder_unzipped / "proteinGroups.txt"
index_cols = [
    "Majority protein IDs",
]
data = pd.read_csv(fpath_proteinGroups, index_col=index_cols, sep="\t")
data.sample(5)

# %% [markdown]
# We mark the protein group by the first protein in the group, ensuring that the protein
# group is still unique.

# %%
new_index = data.index.str.split(";").str[0].rename("first_prot")
assert new_index.is_unique
data = data.reset_index()
data.index = new_index
data

# %% [markdown]
# Get ride of potential contaminants, reverse (decoys) and identified only by a
# modification site
# reference:
# - [cox-labs.github.io/coxdocs/output_tables.html#protein-groups](https://cox-labs.github.io/coxdocs/output_tables.html#protein-groups)

# %%
filters = ["Reverse", "Only identified by site", "Contaminant"]
data[filters].describe()

# %%
mask = data[filters].isna().all(axis=1)
data = data.loc[mask]
data

# %% [markdown]
# Then we can filter the columns that contain the string `LFQ intensity`. The sample names
# are part of the column names (here: `LFQ intensity {sample_name}`)

# %%
stub_intensity = "LFQ intensity"
pgs = data.filter(like=stub_intensity)
pgs


# %% [markdown]
# The associated metadata for protein groups we will keep for reference:

# %%
meta_pgs = data.drop(pgs.columns, axis=1)
meta_pgs

# %% [markdown]
# No we can get rid of the common part `LFQ intensity` and keep only the sample names

# %%
pgs.columns = pgs.columns.str.replace(stub_intensity, "").str.strip()
pgs.columns.name = "sample"
pgs

# %% [markdown]
# ## Parse metadata from column names
# The group could be defined in a sample metadata file, but here we just parse it from the
# sample names by omitting the numbers at the end of the sample name.

# %%
pgs.columns.str.replace(r"\d", "", regex=True)

# %% [markdown]
# We add to the information as a MultiIndex of group and sample name to the columns
# (sample metadata).

# %%
pgs.columns = pd.MultiIndex.from_arrays(
    [pgs.columns.str.replace(r"\d", "", regex=True), pgs.columns],
    names=["group", pgs.columns.name],
)
pgs

# %% [markdown]
# ## Long format and log2 transformation
# From here we can stack both levels, name the values intensity. If we reset the index we
# get the original CKG format.

# %%
pgs = pgs.stack([0, 1]).to_frame("intensity")
pgs

# %% [markdown]
# First we  `log2` transform the data. We first set the zeros to `np.nan` to avoid
# `-inf` values.

# %%
pgs = np.log2(pgs.replace(0.0, np.nan).dropna())
pgs


# %% [markdown]
# Data to be saved in the CKG format: Reset the index.

# %%
pgs.reset_index()

# %% [markdown]
# Done.



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: tags,-all
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Normalization of samples in a dataset example
#
# We will explore an Alzheimer dataset where the data was collected in four different sites.
# We will see that the sites have a an effect where the data is in principal component space
# and in UMAP space. We will then normalize the data and see how the effect on these plots.
#
# Refers to the `acore.normalization` module.

# %% tags=["hide-output"]
# %pip install acore

# %% tags=["hide-input"]
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sklearn
import sklearn.impute
import sklearn.preprocessing
import vuecore.decomposition

import acore.decomposition
import acore.normalization
import acore.sklearn


def plot_umap(X_scaled, y, meta_column, random_state=42) -> plt.Axes:
    """Fit and plot UMAP embedding with two components with colors defined by meta_column."""
    embedding = acore.decomposition.umap.run_umap(
        X_scaled, y, random_state=random_state
    )
    ax = embedding.plot.scatter("UMAP 1", "UMAP 2", c=meta_column, cmap="Paired")
    return ax


def standard_normalize(X: pd.DataFrame) -> pd.DataFrame:
    """Standard normalize data and keep indices of DataFrame."""
    X_scaled = (
        sklearn.preprocessing.StandardScaler()
        .set_output(transform="pandas")
        .fit_transform(X)
    )
    return X_scaled


def median_impute(X: pd.DataFrame) -> pd.DataFrame:
    X_imputed = (
        sklearn.impute.SimpleImputer(strategy="median")
        .set_output(transform="pandas")
        .fit_transform(X)
    )
    return X_imputed


def run_and_plot_pca(
    X_scaled,
    y,
    meta_column: Optional[str] = None,
    n_components: int = 4,
) -> tuple[pd.DataFrame, plt.Figure]:
    PCs, _ = acore.decomposition.pca.run_pca(X_scaled, n_components=n_components)
    PCs.columns = [s.replace("principal component", "PC") for s in PCs.columns]
    fig = vuecore.decomposition.pca_grid(
        PCs=PCs, meta_column=y, n_components=n_components, meta_col_name=meta_column
    )
    return PCs, fig


# %% [markdown]
#
# ## Set some parameters

# %% tags=["parameters"]
fname_metadata: str = (
    "https://raw.githubusercontent.com/RasmussenLab/"
    "njab/HEAD/docs/tutorial/data/alzheimer/meta.csv"  # clincial data
)
fname_omics: str = (
    "https://raw.githubusercontent.com/RasmussenLab/"
    "njab/HEAD/docs/tutorial/data/alzheimer/proteome.csv"  # omics data
)
METACOL: str = "_collection site"  # target column in fname_metadata dataset (binary)
METACOL_LABEL: Optional[str] = "site"  # optional: rename target variable
n_features_max: int = 5
freq_cutoff: float = 0.5  # Omics cutoff for sample completeness
VAL_IDS: str = ""  #
VAL_IDS_query: str = ""
weights: bool = True
FOLDER = "alzheimer"
model_name = "all"

# %% [markdown]
# ## Setup

# %% [markdown]
# ### Load proteomics (protein groups) data

# %%
if METACOL_LABEL is None:
    METACOL_LABEL = METACOL
metadata = (
    pd.read_csv(fname_metadata, usecols=["Sample ID", METACOL], index_col=0)
    .convert_dtypes()
    .rename(columns={METACOL: METACOL_LABEL})
)
omics = pd.read_csv(fname_omics, index_col=0)

# %% [markdown]
# Data shapes

# %%
omics.shape, metadata.shape

# %% [markdown]
# See how common omics features are and remove feature below choosen frequency cutoff

# %%
ax = omics.notna().sum().sort_values().plot(rot=90)

# %% tags=["hide-input"]
M_before = omics.shape[1]
omics = omics.dropna(thresh=int(len(omics) * freq_cutoff), axis=1)
M_after = omics.shape[1]
msg = (
    f"Removed {M_before-M_after} features with more "
    f"than {freq_cutoff*100}% missing values."
    f"\nRemaining features: {M_after} (of {M_before})"
)
print(msg)
# keep a map of all proteins in protein group, but only display first protein
# proteins are unique to protein groups
pg_map = {k: k.split(";")[0] for k in omics.columns}
omics = omics.rename(columns=pg_map)
# log2 transform raw intensity data:
omics = np.log2(omics + 1)
ax = (
    omics.notna()
    .sum()
    .sort_values()
    .plot(
        rot=90,
        ylabel="Number of samples",
        xlabel="Proteins (ranked by missing values)",
    )
)
omics

# %% [markdown]
# ### Sample metadata

# %%
metadata

# %% [markdown]
# Tabulate selected metadata and check for missing values

# %% tags=["hide-input"]
metadata[METACOL_LABEL].value_counts(dropna=False)

# %% tags=["hide-input"]
target_counts = metadata[METACOL_LABEL].value_counts()

if target_counts.sum() < len(metadata):
    print(
        "Target has missing values."
        f" Can only use {target_counts.sum()} of {len(metadata)} samples."
    )
    mask = metadata[METACOL_LABEL].notna()
    metadata, omics = metadata.loc[mask], omics.loc[mask]

if METACOL_LABEL is None:
    METACOL_LABEL = METACOL_LABEL
y = metadata[METACOL_LABEL].astype("category")

# %% [markdown]
# For simplicity we normalize here all samples together, but normally you would need to
# apply the normalization from you training data to the test data. So see these examples
# here as a way to do it for your training data.

# %% [markdown]
# ### Fill missing values for preliminary plots

# %% [markdown]
# Impute using median to impute (before scaling, which can be changed).

# %% tags=["hide-input"]
omics_imputed = median_impute(omics)
assert omics_imputed.isna().sum().sum() == 0
omics_imputed.shape

# %% [markdown]
# Explained variance by first four principal components in data.

# %% tags=["hide-input"]
PCs, pca = acore.decomposition.pca.run_pca(omics_imputed, n_components=4)
ax = vuecore.decomposition.plot_explained_variance(pca)
ax.locator_params(axis="x", integer=True)

# %% [markdown]
# ## Normalization of samples in a dataset
# We will use the `acore.normalization` module to normalize the data.
#
# We will do it for each of the data on the omics dataset which is log transformed,
# but not yet imputed and normalized. Then we will reapply standard
# normalization before replotting the PCA and UMAP plots. The execption is combat as it
# need complete data.

# %% tags=["hide-input"]
omics

# %% [markdown]
# ## Combat normalization
# Correct for batch effects in the data using a robust regression approach normalizing
# mean and scale effetcs out for each feature by batch. Assumes normally distributed data.
#
# > ⚠️ Combat needs imputed data

# %%
# %%time
X = median_impute(omics)
X = acore.normalization.combat_batch_correction(
    X.join(y),
    batch_col="site",
)
X

# %% tags=["hide-input"]
omics_imp = median_impute(X)
omics_imp_scaled = standard_normalize(omics_imp)
PCs, fig = run_and_plot_pca(omics_imp_scaled, y, METACOL_LABEL, n_components=4)
ax = plot_umap(omics_imp_scaled, y, METACOL_LABEL)

# %% [markdown]
# See change by substracting combat normalized data from original data.

# %% tags=["hide-input"]
omics - X


# %% [markdown]
# ## Median normalization
# Substracts a constant from all features of a sample. All samples will have the same
# global median.

# %%
# %%time
X = acore.normalization.normalize_data(omics, "median")
X

# %% tags=["hide-input"]
omics_imp = median_impute(X)
omics_imp_scaled = standard_normalize(omics_imp)
PCs, fig = run_and_plot_pca(omics_imp_scaled, y, METACOL_LABEL, n_components=4)
ax = plot_umap(omics_imp_scaled, y, METACOL_LABEL)

# %% [markdown]
# See change by substracting median normalized data from original data.

# %% tags=["hide-input"]
omics - X

# %%
# %% [markdown]
# ## Z-score normalization
# Normalize a sample by it's mean and standard deviation.

# %%
# %%time
X = acore.normalization.normalize_data(omics, "zscore")
X

# %% tags=["hide-input"]
omics_imp = median_impute(X)
omics_imp_scaled = standard_normalize(omics_imp)
PCs, fig = run_and_plot_pca(omics_imp_scaled, y, METACOL_LABEL, n_components=4)
ax = plot_umap(omics_imp_scaled, y, METACOL_LABEL)

# %% [markdown]
# See change by substracting z-score normalized data from original data.

# %% tags=["hide-input"]
omics_imp_scaled - X

# %% [markdown]
# ## Median Polish Normalization
# - normalize iteratively features and samples to have zero median.

# %%
# %%time
X = acore.normalization.normalize_data(omics, "median_polish")
X

# %% tags=["hide-input"]
omics_imp = median_impute(X)
omics_imp_scaled = standard_normalize(omics_imp)
PCs, fig = run_and_plot_pca(omics_imp_scaled, y, METACOL_LABEL, n_components=4)
ax = plot_umap(omics_imp_scaled, y, METACOL_LABEL)

# %% [markdown]
# See change by substracting median polish normalized data from original data.

# %% tags=["hide-input"]
omics_imp_scaled - X

# %% [markdown]
# ## Quantile normalization
# quantile normalize each feature column.

# %%
# %%time
X = acore.normalization.normalize_data(omics, "quantile")
X

# %% tags=["hide-input"]
omics_imp = median_impute(X)
omics_imp_scaled = standard_normalize(omics_imp)
PCs, fig = run_and_plot_pca(omics_imp_scaled, y, METACOL_LABEL, n_components=4)
ax = plot_umap(omics_imp_scaled, y, METACOL_LABEL)

# %%
omics - X


# %% [markdown]
# ## Linear normalization

# %%
# %%time
X = acore.normalization.normalize_data(omics, "linear")
X

# %% tags=["hide-input"]
omics_imp = median_impute(X)
omics_imp_scaled = standard_normalize(omics_imp)
PCs, fig = run_and_plot_pca(omics_imp_scaled, y, METACOL_LABEL, n_components=4)
ax = plot_umap(omics_imp_scaled, y, METACOL_LABEL)

# %% tags=["hide-input"]
omics - X

# %% [markdown]
# ## Summmary
# Besides the median polish normalization, the structure of the data is not changed
# too much by the normalization using this Alzheimer example. This notebook can be opened
# on colab and might be a good starting point for investigating the effect of normalization
# on your data - or to disect some approaches further.
